# pylint: disable=duplicate-code
"""3-dimensional constructors."""

__all__: list[str] = []

from typing import Any

from .base import AbstractAcceleration3D, AbstractPosition3D, AbstractVelocity3D
from .cartesian import CartesianAcceleration3D, CartesianPosition3D, CartesianVelocity3D

#####################################################################


@AbstractPosition3D.constructor._f.dispatch(precedence=-1)  # noqa: SLF001
def constructor(cls: type[AbstractPosition3D], obj: Any, /) -> CartesianPosition3D:
    """Try to construct a 3D Cartesian position from an object.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> x = Quantity([1, 2, 3], "km")
    >>> cx.AbstractPosition3D.constructor(x)
    CartesianPosition3D(
      x=Quantity[...](value=f32[], unit=Unit("km")),
      y=Quantity[...](value=f32[], unit=Unit("km")),
      z=Quantity[...](value=f32[], unit=Unit("km"))
    )

    """
    return (
        obj
        if isinstance(obj, CartesianPosition3D)
        else CartesianPosition3D.constructor(obj)
    )


@AbstractPosition3D.constructor._f.dispatch(precedence=1)  # noqa: SLF001
def constructor(
    cls: type[AbstractPosition3D], obj: AbstractPosition3D, /
) -> AbstractPosition3D:
    """Construct from a 3D position.

    Examples
    --------
    >>> import coordinax as cx

    >>> cart = cx.CartesianPosition3D.constructor([1, 2, 3], "km")
    >>> cx.AbstractPosition3D.constructor(cart) is cart
    True

    >>> sph = cart.represent_as(cx.SphericalPosition)
    >>> cx.AbstractPosition3D.constructor(sph) is sph
    True

    >>> cyl = cart.represent_as(cx.CylindricalPosition)
    >>> cx.AbstractPosition3D.constructor(cyl) is cyl
    True

    """
    return obj


#####################################################################


@AbstractVelocity3D.constructor._f.dispatch(precedence=-1)  # noqa: SLF001
def constructor(cls: type[AbstractVelocity3D], obj: Any, /) -> CartesianVelocity3D:
    """Try to construct a 3D Cartesian velocity from an object.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> x = Quantity([1, 2, 3], "km / s")
    >>> cx.AbstractVelocity3D.constructor(x)
    CartesianVelocity3D(
      d_x=Quantity[...]( value=f32[], unit=Unit("km / s") ),
      d_y=Quantity[...]( value=f32[], unit=Unit("km / s") ),
      d_z=Quantity[...]( value=f32[], unit=Unit("km / s") )
    )

    """
    return (
        obj
        if isinstance(obj, CartesianVelocity3D)
        else CartesianVelocity3D.constructor(obj)
    )


@AbstractVelocity3D.constructor._f.dispatch(precedence=1)  # noqa: SLF001
def constructor(
    cls: type[AbstractVelocity3D], obj: AbstractVelocity3D, /
) -> AbstractVelocity3D:
    """Construct from a 3D velocity.

    Examples
    --------
    >>> import coordinax as cx

    >>> q = cx.CartesianPosition3D.constructor([1, 1, 1], "km")

    >>> cart = cx.CartesianVelocity3D.constructor([1, 2, 3], "km/s")
    >>> cx.AbstractVelocity3D.constructor(cart) is cart
    True

    >>> sph = cart.represent_as(cx.SphericalVelocity, q)
    >>> cx.AbstractVelocity3D.constructor(sph) is sph
    True

    >>> cyl = cart.represent_as(cx.CylindricalVelocity, q)
    >>> cx.AbstractVelocity3D.constructor(cyl) is cyl
    True

    """
    return obj


#####################################################################


@AbstractAcceleration3D.constructor._f.dispatch(precedence=-1)  # noqa: SLF001
def constructor(
    cls: type[AbstractAcceleration3D], obj: Any, /
) -> CartesianAcceleration3D:
    """Try to construct a 3D Cartesian velocity from an object.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> x = Quantity([1, 2, 3], "km / s2")
    >>> cx.AbstractAcceleration3D.constructor(x)
    CartesianAcceleration3D(
      d2_x=Quantity[...](value=f32[], unit=Unit("km / s2")),
      d2_y=Quantity[...](value=f32[], unit=Unit("km / s2")),
      d2_z=Quantity[...](value=f32[], unit=Unit("km / s2"))
    )

    """
    return (
        obj
        if isinstance(obj, CartesianAcceleration3D)
        else CartesianAcceleration3D.constructor(obj)
    )


@AbstractAcceleration3D.constructor._f.dispatch(precedence=1)  # noqa: SLF001
def constructor(
    cls: type[AbstractAcceleration3D], obj: AbstractAcceleration3D, /
) -> AbstractAcceleration3D:
    """Construct from a 3D velocity.

    Examples
    --------
    >>> import coordinax as cx

    >>> q = cx.CartesianPosition3D.constructor([1, 1, 1], "km")
    >>> p = cx.CartesianVelocity3D.constructor([1, 1, 1], "km/s")

    >>> cart = cx.CartesianAcceleration3D.constructor([1, 2, 3], "km/s2")
    >>> cx.AbstractAcceleration3D.constructor(cart) is cart
    True

    >>> sph = cart.represent_as(cx.SphericalAcceleration, p, q)
    >>> cx.AbstractAcceleration3D.constructor(sph) is sph
    True

    >>> cyl = cart.represent_as(cx.CylindricalAcceleration, p, q)
    >>> cx.AbstractAcceleration3D.constructor(cyl) is cyl
    True

    """
    return obj



"""Representation of coordinates in different systems."""

__all__: list[str] = []

from typing import Any

from plum import dispatch

import quaxed.numpy as xp

from .base import AbstractPosition2D, AbstractVelocity2D
from .cartesian import CartesianAcceleration2D, CartesianPosition2D, CartesianVelocity2D
from .polar import PolarPosition, PolarVelocity
from coordinax._src.base import AbstractPosition


@dispatch
def represent_as(
    current: AbstractPosition2D, target: type[AbstractPosition2D], /, **kwargs: Any
) -> AbstractPosition2D:
    """AbstractPosition2D -> Cartesian2D -> AbstractPosition2D.

    This is the base case for the transformation of 2D vectors.
    """
    return represent_as(represent_as(current, CartesianPosition2D), target)


@dispatch.multi(
    (CartesianPosition2D, type[CartesianPosition2D]),
    (PolarPosition, type[PolarPosition]),
)
def represent_as(
    current: AbstractPosition2D, target: type[AbstractPosition2D], /, **kwargs: Any
) -> AbstractPosition2D:
    """Self transform of 2D vectors."""
    return current


@dispatch.multi(
    (CartesianVelocity2D, type[CartesianVelocity2D], AbstractPosition),
    (PolarVelocity, type[PolarVelocity], AbstractPosition),
)
def represent_as(
    current: AbstractVelocity2D,
    target: type[AbstractVelocity2D],
    position: AbstractPosition,
    /,
    **kwargs: Any,
) -> AbstractVelocity2D:
    """Self transform of 2D Differentials."""
    return current


# =============================================================================
# CartesianPosition2D


@dispatch
def represent_as(
    current: CartesianPosition2D, target: type[PolarPosition], /, **kwargs: Any
) -> PolarPosition:
    """CartesianPosition2D -> PolarPosition.

    The `x` and `y` coordinates are converted to the radial coordinate `r` and
    the angular coordinate `phi`.
    """
    r = xp.sqrt(current.x**2 + current.y**2)
    phi = xp.atan2(current.y, current.x)
    return target(r=r, phi=phi)


# =============================================================================
# PolarPosition


@dispatch
def represent_as(
    current: PolarPosition, target: type[CartesianPosition2D], /, **kwargs: Any
) -> CartesianPosition2D:
    """PolarPosition -> CartesianPosition2D."""
    x = current.r.distance * xp.cos(current.phi)
    y = current.r.distance * xp.sin(current.phi)
    return target(x=x, y=y)


# =============================================================================
# CartesianVelocity2D


@dispatch
def represent_as(
    current: CartesianVelocity2D, target: type[CartesianVelocity2D], /
) -> CartesianVelocity2D:
    """CartesianVelocity2D -> CartesianVelocity2D with no position.

    Cartesian coordinates are an affine coordinate system and so the
    transformation of an n-th order derivative vector in this system do not
    require lower-order derivatives to be specified. See
    https://en.wikipedia.org/wiki/Tensors_in_curvilinear_coordinates for more
    information. This mixin provides a corresponding implementation of the
    `coordinax.represent_as` method for Cartesian velocities.

    Examples
    --------
    >>> import coordinax as cx
    >>> v = cx.CartesianVelocity2D.constructor([1, 1], "m/s")
    >>> cx.represent_as(v, cx.CartesianVelocity2D) is v
    True

    """
    return current


# =============================================================================
# CartesianAcceleration2D


@dispatch
def represent_as(
    current: CartesianAcceleration2D, target: type[CartesianAcceleration2D], /
) -> CartesianAcceleration2D:
    """CartesianAcceleration2D -> CartesianAcceleration2D with no position.

    Cartesian coordinates are an affine coordinate system and so the
    transformation of an n-th order derivative vector in this system do not
    require lower-order derivatives to be specified. See
    https://en.wikipedia.org/wiki/Tensors_in_curvilinear_coordinates for more
    information. This mixin provides a corresponding implementation of the
    `coordinax.represent_as` method for Cartesian vectors.

    Examples
    --------
    >>> import coordinax as cx
    >>> a = cx.CartesianAcceleration2D.constructor([1, 1], "m/s2")
    >>> cx.represent_as(a, cx.CartesianAcceleration2D) is a
    True

    """
    return current



"""Representation of coordinates in different systems."""

__all__ = ["AbstractPosition1D", "AbstractVelocity1D", "AbstractAcceleration1D"]


from abc import abstractmethod
from dataclasses import fields

from jaxtyping import Shaped

import quaxed.numpy as jnp
from unxt import Quantity

from coordinax._src.base import (
    AbstractAcceleration,
    AbstractPosition,
    AbstractVector,
    AbstractVelocity,
)
from coordinax._src.utils import classproperty

#####################################################################


class AbstractPosition1D(AbstractPosition):
    """Abstract representation of 1D coordinates in different systems."""

    @classproperty
    @classmethod
    def _cartesian_cls(cls) -> type[AbstractVector]:
        from .cartesian import CartesianPosition1D

        return CartesianPosition1D

    @classproperty
    @classmethod
    @abstractmethod
    def differential_cls(cls) -> type["AbstractVelocity1D"]:
        raise NotImplementedError


# -------------------------------------------------------------------


@AbstractPosition1D.constructor._f.dispatch  # type: ignore[attr-defined, misc]  # noqa: SLF001
def constructor(
    cls: type[AbstractPosition1D],
    obj: Shaped[Quantity["length"], "*batch"] | Shaped[Quantity["length"], "*batch 1"],
    /,
) -> AbstractPosition1D:
    """Construct a 1D position.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> cx.CartesianPosition1D.constructor(Quantity(1, "meter"))
    CartesianPosition1D( x=Quantity[...](value=f32[], unit=Unit("m")) )

    >>> cx.CartesianPosition1D.constructor(Quantity([1], "meter"))
    CartesianPosition1D( x=Quantity[...](value=f32[], unit=Unit("m")) )

    >>> cx.RadialPosition.constructor(Quantity(1, "meter"))
    RadialPosition(r=Distance(value=f32[], unit=Unit("m")))

    >>> cx.RadialPosition.constructor(Quantity([1], "meter"))
    RadialPosition(r=Distance(value=f32[], unit=Unit("m")))

    """
    comps = {f.name: jnp.atleast_1d(obj)[..., i] for i, f in enumerate(fields(cls))}
    return cls(**comps)


#####################################################################


class AbstractVelocity1D(AbstractVelocity):
    """Abstract representation of 1D differentials in different systems."""

    @classproperty
    @classmethod
    def _cartesian_cls(cls) -> type[AbstractVector]:
        from .cartesian import CartesianVelocity1D

        return CartesianVelocity1D

    @classproperty
    @classmethod
    @abstractmethod
    def integral_cls(cls) -> type[AbstractPosition1D]:
        raise NotImplementedError

    @classproperty
    @classmethod
    @abstractmethod
    def differential_cls(cls) -> type[AbstractAcceleration]:
        raise NotImplementedError


# -------------------------------------------------------------------


@AbstractVelocity1D.constructor._f.dispatch  # type: ignore[attr-defined, misc]  # noqa: SLF001
def constructor(
    cls: type[AbstractVelocity1D],
    obj: Shaped[Quantity["speed"], "*batch"] | Shaped[Quantity["speed"], "*batch 1"],
    /,
) -> AbstractVelocity1D:
    """Construct a 1D velocity.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> cx.CartesianVelocity1D.constructor(Quantity(1, "m/s"))
    CartesianVelocity1D( d_x=Quantity[...]( value=...i32[], unit=Unit("m / s") ) )

    >>> cx.CartesianVelocity1D.constructor(Quantity([1], "m/s"))
    CartesianVelocity1D( d_x=Quantity[...]( value=i32[], unit=Unit("m / s") ) )

    >>> cx.RadialVelocity.constructor(Quantity(1, "m/s"))
    RadialVelocity( d_r=Quantity[...]( value=...i32[], unit=Unit("m / s") ) )

    >>> cx.RadialVelocity.constructor(Quantity([1], "m/s"))
    RadialVelocity( d_r=Quantity[...]( value=i32[], unit=Unit("m / s") ) )

    """
    comps = {f.name: jnp.atleast_1d(obj)[..., i] for i, f in enumerate(fields(cls))}
    return cls(**comps)


#####################################################################


class AbstractAcceleration1D(AbstractAcceleration):
    """Abstract representation of 1D acceleration in different systems."""

    @classproperty
    @classmethod
    def _cartesian_cls(cls) -> type[AbstractVector]:
        from .cartesian import CartesianAcceleration1D

        return CartesianAcceleration1D

    @classproperty
    @classmethod
    @abstractmethod
    def integral_cls(cls) -> type[AbstractVelocity1D]:
        raise NotImplementedError


# -------------------------------------------------------------------


@AbstractAcceleration1D.constructor._f.dispatch  # type: ignore[attr-defined, misc]  # noqa: SLF001
def constructor(
    cls: type[AbstractAcceleration1D],
    obj: Shaped[Quantity["acceleration"], "*batch"]
    | Shaped[Quantity["acceleration"], "*batch 1"],
    /,
) -> AbstractAcceleration1D:
    """Construct a 1D acceleration.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> cx.CartesianAcceleration1D.constructor(Quantity(1, "m/s2"))
    CartesianAcceleration1D( d2_x=... )

    >>> cx.CartesianAcceleration1D.constructor(Quantity([1], "m/s2"))
    CartesianAcceleration1D( d2_x=Quantity[...](value=i32[], unit=Unit("m / s2")) )

    >>> cx.RadialAcceleration.constructor(Quantity(1, "m/s2"))
    RadialAcceleration( d2_r=... )

    >>> cx.RadialAcceleration.constructor(Quantity([1], "m/s2"))
    RadialAcceleration( d2_r=Quantity[...](value=i32[], unit=Unit("m / s2")) )

    """
    comps = {f.name: jnp.atleast_1d(obj)[..., i] for i, f in enumerate(fields(cls))}
    return cls(**comps)



# ruff: noqa: ERA001
"""Galilean coordinate transformations."""

__all__ = ["GalileanOperator"]


from typing import TYPE_CHECKING, Any, final, overload

import equinox as eqx

import quaxed.numpy as xp
from dataclassish.converters import Unless
from unxt import Quantity

from .base import AbstractGalileanOperator
from .boost import GalileanBoostOperator
from .rotation import GalileanRotationOperator
from .translation import GalileanTranslationOperator
from coordinax._src.operators.base import AbstractOperator
from coordinax._src.operators.composite import AbstractCompositeOperator
from coordinax._src.operators.funcs import simplify_op
from coordinax._src.operators.identity import IdentityOperator
from coordinax._src.operators.sequential import OperatorSequence

if TYPE_CHECKING:
    from typing_extensions import Self


@final
class GalileanOperator(AbstractCompositeOperator, AbstractGalileanOperator):
    r"""Operator for general Galilean transformations.

    In the transformed frame the coordinates are given by:

    .. math::

        (t,\mathbf{x}) \mapsto (t+s, R \mathbf{x} + \mathbf{v} t + \mathbf{a})

    where :math:`R` is the rotation matrix, :math:`\mathbf{v}` is the boost
    velocity, :math:`\mathbf{a}` is the spatial translationl, and :math:`s` is
    the time translation. This is equivalent to a sequential operation of 1. a
    rotation, 2. a translation, 3. a boost.

    Parameters
    ----------
    translation : `coordinax.operators.GalileanTranslationOperator`
        The spatial translation of the frame. See
        :class:`coordinax.operators.GalileanTranslationOperator` for
        alternative inputs to construct this parameter.
    velocity : :class:`coordinax.operators.GalileanBoostOperator`
        The boost to the frame. See
        :class:`coordinax.operators.GalileanBoostOperator` for
        alternative inputs to construct this parameter.

    Examples
    --------
    We start with the required imports:

    >>> from unxt import Quantity
    >>> import coordinax as cx

    We can then create a Galilean operator:

    >>> op = cx.operators.GalileanOperator(
    ...     translation=Quantity([0., 2., 3., 4.], "kpc"),
    ...     velocity=Quantity([1., 2., 3.], "km/s"))
    >>> op
    GalileanOperator(
      rotation=GalileanRotationOperator(rotation=f32[3,3]),
      translation=GalileanTranslationOperator(
        translation=FourVector(
          t=Quantity[PhysicalType('time')](value=f32[], unit=Unit("kpc s / km")),
          q=CartesianPosition3D( ... ) )
      ),
      velocity=GalileanBoostOperator( velocity=CartesianVelocity3D( ... ) )
    )

    Note that the translation is a
    :class:`coordinax.operators.GalileanTranslationOperator` with a
    :class:`vector.FourVector` translation, and the velocity is a
    :class:`coordinax.operators.GalileanBoostOperator` with a
    :class:`vector.CartesianVelocity3D` velocity. We can also construct them
    directly, which allows for other vector types.

    >>> op = cx.operators.GalileanOperator(
    ...     translation=cx.operators.GalileanTranslationOperator(
    ...         cx.FourVector(t=Quantity(2.5, "Gyr"),
    ...                    q=cx.SphericalPosition(r=Quantity(1, "kpc"),
    ...                                         theta=Quantity(90, "deg"),
    ...                                         phi=Quantity(0, "rad") ) ) ),
    ...     velocity=cx.operators.GalileanBoostOperator(
    ...         cx.CartesianVelocity3D(d_x=Quantity(1, "km/s"),
    ...                                    d_y=Quantity(2, "km/s"),
    ...                                    d_z=Quantity(3, "km/s")))
    ... )
    >>> op
    GalileanOperator(
      rotation=GalileanRotationOperator(rotation=f32[3,3]),
      translation=GalileanTranslationOperator(
        translation=FourVector(
          t=Quantity[PhysicalType('time')](value=f32[], unit=Unit("Gyr")),
          q=SphericalPosition( ... )
        )
      ),
      velocity=GalileanBoostOperator( velocity=CartesianVelocity3D( ... ) )
    )

    Galilean operators can be applied to :class:`vector.FourVector`:

    >>> w = cx.FourVector.constructor([0, 0, 0, 0], "kpc")
    >>> new = op(w)
    >>> new
    FourVector(
      t=Quantity[PhysicalType('time')](value=f32[], unit=Unit("kpc s / km")),
      q=CartesianPosition3D( ... )
    )
    >>> new.t.to_units("Gyr").value.round(2)
    Array(2.5, dtype=float32)
    >>> new.q.x
    Quantity['length'](Array(3.5567803, dtype=float32), unit='kpc')

    Also the Galilean operators can also be applied to
    :class:`vector.AbstractPosition3D` and :class:`unxt.Quantity`:

    >>> q = cx.CartesianPosition3D.constructor([0, 0, 0], "kpc")
    >>> t = Quantity(0, "Gyr")
    >>> newq, newt = op(q, t)
    >>> newq.x
    Quantity['length'](Array(3.5567803, dtype=float32), unit='kpc')
    >>> newt
    Quantity['time'](Array(2.5, dtype=float32), unit='Gyr')

    """

    rotation: GalileanRotationOperator = eqx.field(
        default=GalileanRotationOperator(xp.eye(3)),
        converter=GalileanRotationOperator,
    )
    """The in-frame spatial rotation."""

    translation: GalileanTranslationOperator = eqx.field(
        default=GalileanTranslationOperator(Quantity([0, 0, 0, 0], "kpc")),
        converter=Unless(
            GalileanTranslationOperator, converter=GalileanTranslationOperator
        ),
    )
    """The temporal + spatial translation.

    The translation vector [T, Q].  This parameters accetps either a
    :class:`coordinax.operators.GalileanTranslationOperator` instance or
    any input that can be used to construct a :meth:`vector.FourVector`, using
    :meth:`vector.FourVector.constructor`. See :class:`vector.FourVector` for
    details.
    """

    velocity: GalileanBoostOperator = eqx.field(
        default=GalileanBoostOperator(Quantity([0, 0, 0], "km/s")),
        converter=Unless(GalileanBoostOperator, converter=GalileanBoostOperator),
    )
    """The boost to the frame.

    This parameters accepts either a
    :class:`coordinax.operators.GalileanBoostOperator` instance or any
    input that can be used to construct a
    :class:`vector.CartesianVelocity3D`, using
    :meth:`vector.CartesianVelocity3D.constructor`. See
    :class:`vector.CartesianVelocity3D` for details.
    """

    @property
    def operators(
        self,
    ) -> tuple[
        GalileanRotationOperator, GalileanTranslationOperator, GalileanBoostOperator
    ]:
        """Rotation -> translateion -> boost."""
        return (self.rotation, self.translation, self.velocity)

    @overload
    def __getitem__(self, key: int) -> AbstractOperator: ...

    @overload
    def __getitem__(self, key: slice) -> "Self": ...

    def __getitem__(self, key: int | slice) -> "AbstractOperator | Self":
        if isinstance(key, int):
            return self.operators[key]
        return OperatorSequence(self.operators[key])


@simplify_op.register
def _simplify_op_galilean(op: GalileanOperator, /, **kwargs: Any) -> AbstractOperator:
    """Simplify a Galilean operator."""
    # Check if all the sub-operators can be simplified to the identity.
    if all(
        isinstance(simplify_op(x, **kwargs), IdentityOperator) for x in op.operators
    ):
        return IdentityOperator()

    return op



# ruff: noqa: ERA001
"""Galilean coordinate transformations."""

__all__ = ["GalileanRotationOperator"]


from collections.abc import Mapping
from dataclasses import replace
from typing import Any, Literal, final

import equinox as eqx
import jax
from jaxtyping import Array, Shaped
from plum import convert
from quax import quaxify

import quaxed.numpy as jnp
from unxt import Quantity, ustrip

from .base import AbstractGalileanOperator
from coordinax._src.base import ToUnitsOptions
from coordinax._src.d3.base import AbstractPosition3D
from coordinax._src.d3.cartesian import CartesianPosition3D
from coordinax._src.operators.base import AbstractOperator, op_call_dispatch
from coordinax._src.operators.funcs import simplify_op
from coordinax._src.operators.identity import IdentityOperator

vec_matmul = quaxify(jax.numpy.vectorize(jax.numpy.matmul, signature="(3,3),(3)->(3)"))


def converter(x: Any) -> Array:
    """Convert the input to a rotation matrix."""
    if isinstance(x, GalileanRotationOperator):
        out = x.rotation
    elif isinstance(x, Quantity):
        out = ustrip("", x)
    else:
        out = x
    return jnp.asarray(out)


@final
class GalileanRotationOperator(AbstractGalileanOperator):
    r"""Operator for Galilean rotations.

    The coordinate transform is given by:

    .. math::

        (t,\mathbf{x}) \mapsto (t, R \mathbf{x})

    where :math:`R` is the rotation matrix.  Note this is NOT time dependent.

    Parameters
    ----------
    rotation : Array[float, (3, 3)]
        The rotation matrix.

    Raises
    ------
    ValueError
        If the rotation matrix is not orthogonal.

    Notes
    -----
    The Galilean rotation is not a time-dependent transformation.  This is part
    of the inhomogeneous Galilean group, which is the group of transformations
    that leave the space-time interval invariant.

    Examples
    --------
    We start with the required imports:

    >>> import jax.numpy as jnp
    >>> from unxt import Quantity
    >>> import coordinax as cx
    >>> import coordinax.operators as co

    We can then create a rotation operator:

    >>> theta = jnp.pi / 4  # 45 degrees
    >>> Rz = jnp.asarray([[jnp.cos(theta), -jnp.sin(theta), 0],
    ...                   [jnp.sin(theta), jnp.cos(theta),  0],
    ...                   [0,              0,               1]])
    >>> op = co.GalileanRotationOperator(Rz)
    >>> op
    GalileanRotationOperator(rotation=f32[3,3])

    Translation operators can be applied to a Quantity[float, (N, 3), "...]:

    >>> q = Quantity([1, 0, 0], "m")
    >>> t = Quantity(1, "s")
    >>> newq, newt = op(q, t)
    >>> newq
    Quantity['length'](Array([0.70710677, 0.70710677, 0. ], dtype=float32), unit='m')

    The time is not affected by the rotation.

    >>> newt
    Quantity['time'](Array(1, dtype=int32, ...), unit='s')

    This also works for a batch of vectors:

    >>> q = Quantity([[1, 0, 0], [0, 1, 0]], "m")
    >>> t = Quantity(0, "s")

    >>> newq, newt = op(q, t)
    >>> newq
    Quantity['length'](Array([[ 0.70710677,  0.70710677,  0.        ],
                              [-0.70710677,  0.70710677,  0.        ]], dtype=float32),
                       unit='m')

    Translation operators can be applied to :class:`vector.AbstractPosition3D`:

    >>> q = cx.CartesianPosition3D.constructor(q)  # from the previous example
    >>> newq, newt = op(q, t)
    >>> newq.x
    Quantity['length'](Array([ 0.70710677, -0.70710677], dtype=float32), unit='m')
    >>> newq.norm().value.round(2)
    Array([1., 1.], dtype=float32)

    """

    rotation: Shaped[Array, "3 3"] = eqx.field(converter=converter)
    """The rotation vector."""

    check_tol: Mapping[str, Any] = eqx.field(
        default_factory=lambda: {"atol": 1e-7}, repr=False, static=True
    )

    # TODO: fix so that it works in jitted contexts
    # def __check_init__(self) -> None:
    #     # Check that the rotation matrix is orthogonal.
    #     _ = eqx.error_if(
    #         self.rotation,
    #         not jnp.allclose(
    #             self.rotation @ self.rotation.T, jnp.eye(3), **self.check_tol
    #         ),
    #         "The rotation matrix must be orthogonal.",
    #     )

    # -----------------------------------------------------

    @property
    def is_inertial(self) -> Literal[True]:
        """Galilean rotation is an inertial-frame preserving transform.

        Examples
        --------
        >>> import quaxed.numpy as jnp
        >>> from unxt import Quantity
        >>> from coordinax.operators import GalileanRotationOperator

        >>> theta = Quantity(45, "deg")
        >>> Rz = jnp.asarray([[jnp.cos(theta), -jnp.sin(theta), 0],
        ...                  [jnp.sin(theta), jnp.cos(theta),  0],
        ...                  [0,             0,              1]])
        >>> op = GalileanRotationOperator(Rz)
        >>> op.is_inertial
        True

        """
        return True

    @property
    def inverse(self) -> "GalileanRotationOperator":
        """The inverse of the operator.

        Examples
        --------
        >>> import quaxed.numpy as jnp
        >>> from unxt import Quantity
        >>> from coordinax.operators import GalileanRotationOperator

        >>> theta = Quantity(45, "deg")
        >>> Rz = jnp.asarray([[jnp.cos(theta), -jnp.sin(theta), 0],
        ...                  [jnp.sin(theta), jnp.cos(theta),  0],
        ...                  [0,             0,              1]])
        >>> op = GalileanRotationOperator(Rz)
        >>> op.inverse
        GalileanRotationOperator(rotation=f32[3,3])

        >>> jnp.allclose(op.rotation, op.inverse.rotation.T)
        Array(True, dtype=bool)

        """
        return replace(self, rotation=self.rotation.T)

    # -----------------------------------------------------

    @op_call_dispatch(precedence=1)
    def __call__(
        self: "GalileanRotationOperator", q: Shaped[Quantity["length"], "*batch 3"], /
    ) -> Shaped[Quantity["length"], "*batch 3"]:
        """Apply the boost to the coordinates.

        Examples
        --------
        >>> import quaxed.numpy as jnp
        >>> from unxt import Quantity
        >>> from coordinax.operators import GalileanRotationOperator

        >>> theta = Quantity(45, "deg")
        >>> Rz = jnp.asarray([[jnp.cos(theta), -jnp.sin(theta), 0],
        ...                  [jnp.sin(theta), jnp.cos(theta),  0],
        ...                  [0,             0,              1]])
        >>> op = GalileanRotationOperator(Rz)

        >>> q = Quantity([1, 0, 0], "m")
        >>> t = Quantity(1, "s")
        >>> newq, newt = op(q, t)
        >>> newq
        Quantity[...](Array([0.70710677, 0.70710677, 0. ], dtype=float32), unit='m')

        The time is not affected by the rotation.
        >>> newt
        Quantity['time'](Array(1, dtype=int32, ...), unit='s')

        """
        return vec_matmul(self.rotation, q)

    @op_call_dispatch(precedence=1)
    def __call__(
        self: "GalileanRotationOperator", q: AbstractPosition3D, /
    ) -> AbstractPosition3D:
        """Apply the boost to the coordinates.

        Examples
        --------
        >>> import quaxed.numpy as jnp
        >>> from unxt import Quantity
        >>> import coordinax as cx
        >>> from coordinax.operators import GalileanRotationOperator

        >>> theta = Quantity(45, "deg")
        >>> Rz = jnp.asarray([[jnp.cos(theta), -jnp.sin(theta), 0],
        ...                  [jnp.sin(theta), jnp.cos(theta),  0],
        ...                  [0,             0,              1]])
        >>> op = GalileanRotationOperator(Rz)

        >>> q = cx.CartesianPosition3D.constructor([1, 0, 0], "m")
        >>> t = Quantity(1, "s")
        >>> newq, newt = op(q, t)
        >>> newq.x
        Quantity['length'](Array(0.70710677, dtype=float32), unit='m')

        The time is not affected by the rotation.
        >>> newt
        Quantity['time'](Array(1, dtype=int32, ...), unit='s')

        """
        vec = convert(  # Array[float, (N, 3)]
            q.represent_as(CartesianPosition3D).to_units(ToUnitsOptions.consistent),
            Quantity,
        )
        rcart = CartesianPosition3D.constructor(vec_matmul(self.rotation, vec))
        return rcart.represent_as(type(q))

    @op_call_dispatch(precedence=1)
    def __call__(
        self: "GalileanRotationOperator",
        q: AbstractPosition3D,
        t: Quantity["time"],
        /,
    ) -> tuple[AbstractPosition3D, Quantity["time"]]:
        return self(q), t

    @op_call_dispatch(precedence=1)
    def __call__(
        self: "GalileanRotationOperator",
        q: AbstractPosition3D,
        t: Quantity["time"],
        /,
    ) -> tuple[AbstractPosition3D, Quantity["time"]]:
        return self(q), t


@simplify_op.register
def _simplify_op_rotation(
    op: GalileanRotationOperator, /, **kwargs: Any
) -> AbstractOperator:
    if jnp.allclose(op.rotation, jnp.eye(3), **kwargs):
        return IdentityOperator()
    return op



"""Transformations between representations."""

__all__ = ["represent_as"]

from typing import Any
from warnings import warn

import astropy.units as u
from plum import dispatch

import quaxed.numpy as jnp
from unxt import Quantity

from coordinax._src.d1.cartesian import CartesianPosition1D
from coordinax._src.d1.radial import RadialPosition
from coordinax._src.d2.base import AbstractPosition2D
from coordinax._src.d2.cartesian import CartesianPosition2D
from coordinax._src.d2.polar import PolarPosition
from coordinax._src.d3.base import AbstractPosition3D
from coordinax._src.d3.cartesian import CartesianPosition3D
from coordinax._src.d3.cylindrical import CylindricalPosition
from coordinax._src.d3.mathspherical import MathSphericalPosition
from coordinax._src.d3.spherical import SphericalPosition
from coordinax._src.exceptions import IrreversibleDimensionChange


@dispatch.multi(
    (CartesianPosition2D, type[CylindricalPosition]),
    (CartesianPosition2D, type[SphericalPosition]),
    (CartesianPosition2D, type[MathSphericalPosition]),
)
def represent_as(
    current: AbstractPosition2D,
    target: type[AbstractPosition3D],
    /,
    z: Quantity = Quantity(0.0, u.m),
    **kwargs: Any,
) -> AbstractPosition3D:
    """AbstractPosition2D -> Cartesian2D -> Cartesian3D -> AbstractPosition3D.

    The 2D vector is in the xy plane. The `z` coordinate is a keyword argument and
    defaults to 0.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> x = cx.CartesianPosition2D.constructor([1.0, 2.0], "km")

    >>> x2 = cx.represent_as(x, cx.CylindricalPosition, z=Quantity(14, "km"))
    >>> x2
    CylindricalPosition( rho=Quantity[...](value=f32[], unit=Unit("km")),
                       phi=Quantity[...](value=f32[], unit=Unit("rad")),
                       z=Quantity[...](value=f32[], unit=Unit("km")) )
    >>> x2.z
    Quantity['length'](Array(14., dtype=float32), unit='km')

    >>> x3 = cx.represent_as(x, cx.SphericalPosition, z=Quantity(14, "km"))
    >>> x3
    SphericalPosition( r=Distance(value=f32[], unit=Unit("km")),
                     theta=Quantity[...](value=f32[], unit=Unit("rad")),
                     phi=Quantity[...](value=f32[], unit=Unit("rad")) )
    >>> x3.r
    Distance(Array(14.177447, dtype=float32), unit='km')

    >>> x3 = cx.represent_as(x, cx.MathSphericalPosition, z=Quantity(14, "km"))
    >>> x3
    MathSphericalPosition( r=Distance(value=f32[], unit=Unit("km")),
                         theta=Quantity[...](value=f32[], unit=Unit("rad")),
                         phi=Quantity[...](value=f32[], unit=Unit("rad")) )
    >>> x3.r
    Distance(Array(14.177447, dtype=float32), unit='km')

    """
    cart2 = represent_as(current, CartesianPosition2D)
    cart3 = represent_as(cart2, CartesianPosition3D, z=z)
    return represent_as(cart3, target)


@dispatch.multi(
    (PolarPosition, type[CartesianPosition3D]),
)
def represent_as(
    current: AbstractPosition2D,
    target: type[AbstractPosition3D],
    /,
    z: Quantity = Quantity(0.0, u.m),
    **kwargs: Any,
) -> AbstractPosition3D:
    """AbstractPosition2D -> PolarPosition -> Cylindrical -> AbstractPosition3D.

    The 2D vector is in the xy plane. The `z` coordinate is a keyword argument and
    defaults to 0.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> x = cx.PolarPosition(r=Quantity(1.0, "km"), phi=Quantity(10.0, "deg"))

    >>> x2 = cx.represent_as(x, cx.CartesianPosition3D, z=Quantity(14, "km"))
    >>> x2
    CartesianPosition3D( x=Quantity[...](value=f32[], unit=Unit("km")),
                       y=Quantity[...](value=f32[], unit=Unit("km")),
                       z=Quantity[...](value=f32[], unit=Unit("km")) )
    >>> x2.z
    Quantity['length'](Array(14., dtype=float32), unit='km')

    """
    polar = represent_as(current, PolarPosition)
    cyl = represent_as(polar, CylindricalPosition, z=z)
    return represent_as(cyl, target)


# =============================================================================
# CartesianPosition2D


# -----------------------------------------------
# 1D


@dispatch
def represent_as(
    current: CartesianPosition2D, target: type[CartesianPosition1D], /, **kwargs: Any
) -> CartesianPosition1D:
    """CartesianPosition2D -> CartesianPosition1D.

    The `y` coordinate is dropped.

    Examples
    --------
    >>> import warnings
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> x = cx.CartesianPosition2D.constructor([1.0, 2.0], "km")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.CartesianPosition1D, z=Quantity(14, "km"))
    >>> x2
    CartesianPosition1D( x=Quantity[...](value=f32[], unit=Unit("km")) )
    >>> x2.x
    Quantity['length'](Array(1., dtype=float32), unit='km')

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(x=current.x)


@dispatch
def represent_as(
    current: CartesianPosition2D, target: type[RadialPosition], /, **kwargs: Any
) -> RadialPosition:
    """CartesianPosition2D -> RadialPosition.

    The `x` and `y` coordinates are converted to the radial coordinate `r`.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> x = cx.CartesianPosition2D.constructor([1.0, 2.0], "km")

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.RadialPosition, z=Quantity(14, "km"))
    >>> x2
    RadialPosition(r=Distance(value=f32[], unit=Unit("km")))
    >>> x2.r
    Distance(Array(2.236068, dtype=float32), unit='km')

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(r=jnp.sqrt(current.x**2 + current.y**2))


# -----------------------------------------------
# 3D


@dispatch
def represent_as(
    current: CartesianPosition2D,
    target: type[CartesianPosition3D],
    /,
    *,
    z: Quantity = Quantity(0.0, u.m),
    **kwargs: Any,
) -> CartesianPosition3D:
    """CartesianPosition2D -> CartesianPosition3D.

    The `x` and `y` coordinates are converted to the `x` and `y` coordinates of
    the 3D system.  The `z` coordinate is a keyword argument and defaults to 0.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> x = cx.CartesianPosition2D.constructor([1.0, 2.0], "km")

    >>> x2 = cx.represent_as(x, cx.CartesianPosition3D, z=Quantity(14, "km"))
    >>> x2
    CartesianPosition3D( x=Quantity[...](value=f32[], unit=Unit("km")),
                       y=Quantity[...](value=f32[], unit=Unit("km")),
                       z=Quantity[...](value=f32[], unit=Unit("km")) )
    >>> x2.z
    Quantity['length'](Array(14., dtype=float32), unit='km')

    """
    return target(x=current.x, y=current.y, z=z)


# =============================================================================
# PolarPosition

# -----------------------------------------------
# 1D


@dispatch
def represent_as(
    current: PolarPosition, target: type[CartesianPosition1D], /, **kwargs: Any
) -> CartesianPosition1D:
    """PolarPosition -> CartesianPosition1D.

    Examples
    --------
    >>> import warnings
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> x = cx.PolarPosition(r=Quantity(1.0, "km"), phi=Quantity(10.0, "deg"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.CartesianPosition1D)
    >>> x2
    CartesianPosition1D( x=Quantity[...](value=f32[], unit=Unit("km")) )
    >>> x2.x
    Quantity['length'](Array(0.9848077, dtype=float32), unit='km')

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(x=current.r * jnp.cos(current.phi))


@dispatch
def represent_as(
    current: PolarPosition, target: type[RadialPosition], /, **kwargs: Any
) -> RadialPosition:
    """PolarPosition -> RadialPosition.

    Examples
    --------
    >>> import warnings
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> x = cx.PolarPosition(r=Quantity(1.0, "km"), phi=Quantity(10.0, "deg"))

    >>> with warnings.catch_warnings():
    ...     warnings.simplefilter("ignore")
    ...     x2 = cx.represent_as(x, cx.RadialPosition)
    >>> x2
    RadialPosition(r=Distance(value=f32[], unit=Unit("km")))
    >>> x2.r
    Distance(Array(1., dtype=float32), unit='km')

    """
    warn("irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)
    return target(r=current.r)


# -----------------------------------------------
# 3D


@dispatch
def represent_as(
    current: PolarPosition,
    target: type[SphericalPosition],
    /,
    theta: Quantity["angle"] = Quantity(0.0, u.radian),  # type: ignore[name-defined]
    **kwargs: Any,
) -> SphericalPosition:
    """PolarPosition -> SphericalPosition.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> x = cx.PolarPosition(r=Quantity(1.0, "km"), phi=Quantity(10.0, "deg"))

    >>> x2 = cx.represent_as(x, cx.SphericalPosition, theta=Quantity(14, "deg"))
    >>> x2
    SphericalPosition( r=Distance(value=f32[], unit=Unit("km")),
                     theta=Quantity[...](value=f32[], unit=Unit("deg")),
                     phi=Quantity[...](value=f32[], unit=Unit("deg")) )
    >>> x2.theta
    Quantity['angle'](Array(14., dtype=float32), unit='deg')

    """
    return target.constructor(r=current.r, theta=theta, phi=current.phi)


@dispatch
def represent_as(
    current: PolarPosition,
    target: type[MathSphericalPosition],
    /,
    phi: Quantity["angle"] = Quantity(0.0, u.radian),  # type: ignore[name-defined]
    **kwargs: Any,
) -> MathSphericalPosition:
    """PolarPosition -> MathSphericalPosition.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> x = cx.PolarPosition(r=Quantity(1.0, "km"), phi=Quantity(10.0, "deg"))

    >>> x2 = cx.represent_as(x, cx.MathSphericalPosition, phi=Quantity(14, "deg"))
    >>> x2
    MathSphericalPosition( r=Distance(value=f32[], unit=Unit("km")),
                         theta=Quantity[...](value=f32[], unit=Unit("deg")),
                         phi=Quantity[...](value=f32[], unit=Unit("deg")) )
    >>> x2.phi
    Quantity['angle'](Array(14., dtype=float32), unit='deg')

    """
    return target.constructor(r=current.r, phi=phi, theta=current.phi)


@dispatch
def represent_as(
    current: PolarPosition,
    target: type[CylindricalPosition],
    /,
    *,
    z: Quantity["length"] = Quantity(0.0, u.m),  # type: ignore[name-defined]
    **kwargs: Any,
) -> CylindricalPosition:
    """PolarPosition -> CylindricalPosition.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> x = cx.PolarPosition(r=Quantity(1.0, "km"), phi=Quantity(10.0, "deg"))

    >>> x2 = cx.represent_as(x, cx.CylindricalPosition, z=Quantity(14, "km"))
    >>> x2
    CylindricalPosition( rho=Quantity[...](value=f32[], unit=Unit("km")),
                       phi=Quantity[...](value=f32[], unit=Unit("deg")),
                       z=Quantity[...](value=f32[], unit=Unit("km")) )
    >>> x2.z
    Quantity['length'](Array(14., dtype=float32), unit='km')

    """
    return target(rho=current.r, phi=current.phi, z=z)



"""Test using Jax operations."""

import equinox as eqx
import pytest

from dataclassish import field_items
from unxt import AbstractQuantity

import coordinax as cx
from coordinax._src.base.base_pos import POSITION_CLASSES

POSITION_CLASSES_3D = [
    c for c in POSITION_CLASSES if issubclass(c, cx.AbstractPosition3D)
]


# TODO: cycle through all representations
@pytest.fixture(params=POSITION_CLASSES_3D)
def q(request) -> cx.AbstractPosition:
    """Fixture for 3D Vectors."""
    q = cx.CartesianPosition3D.constructor([1, 2, 3], "kpc")
    return q.represent_as(request.param)


@eqx.filter_jit
def func(
    q: cx.AbstractPosition, target: type[cx.AbstractPosition]
) -> cx.AbstractPosition:
    return q.represent_as(target)


@pytest.mark.parametrize("target", POSITION_CLASSES_3D)
def test_jax_through_representation(
    q: cx.AbstractPosition, target: type[cx.AbstractPosition]
) -> None:
    """Test using Jax operations through representation."""
    newq = func(q, target)

    assert isinstance(newq, cx.AbstractPosition)
    for k, f in field_items(newq):
        assert isinstance(f, AbstractQuantity), k



"""Intra-ecosystem Compatibility."""

__all__: list[str] = []


from typing import TypeAlias

from jaxtyping import Shaped
from plum import convert

from unxt import Quantity

from .cartesian import CartesianPosition2D
from coordinax._src.operators.base import AbstractOperator, op_call_dispatch
from coordinax._src.typing import TimeBatchOrScalar

Q2: TypeAlias = Shaped[Quantity["length"], "*#batch 2"]


@op_call_dispatch
def call(self: AbstractOperator, x: Q2, /) -> Q2:
    """Dispatch to the operator's `__call__` method."""
    return self(CartesianPosition2D.constructor(x))


@op_call_dispatch
def call(
    self: AbstractOperator, x: Q2, t: TimeBatchOrScalar, /
) -> tuple[Q2, TimeBatchOrScalar]:
    """Dispatch to the operator's `__call__` method."""
    vec, t = self(CartesianPosition2D.constructor(x), t)
    return convert(vec, Quantity), t



"""Built-in vector classes."""

__all__ = [
    "AbstractSphericalPosition",
    "AbstractSphericalVelocity",
    "AbstractSphericalAcceleration",
]

from abc import abstractmethod
from typing_extensions import override

from unxt import Quantity

from .base import AbstractAcceleration3D, AbstractPosition3D, AbstractVelocity3D
from coordinax._src.utils import classproperty

_90d = Quantity(90, "deg")
_180d = Quantity(180, "deg")
_360d = Quantity(360, "deg")


class AbstractSphericalPosition(AbstractPosition3D):
    """Abstract spherical vector representation."""

    @override
    @classproperty
    @classmethod
    @abstractmethod
    def differential_cls(cls) -> "type[AbstractSphericalVelocity]": ...


class AbstractSphericalVelocity(AbstractVelocity3D):
    """Spherical differential representation."""

    @override
    @classproperty
    @classmethod
    @abstractmethod
    def integral_cls(cls) -> type[AbstractSphericalPosition]: ...

    @override
    @classproperty
    @classmethod
    @abstractmethod
    def differential_cls(cls) -> "type[AbstractSphericalAcceleration]": ...


class AbstractSphericalAcceleration(AbstractAcceleration3D):
    """Spherical acceleration representation."""

    @override
    @classproperty
    @classmethod
    @abstractmethod
    def integral_cls(cls) -> type[AbstractSphericalVelocity]: ...



# ruff: noqa: ERA001
"""Galilean coordinate transformations."""

__all__ = ["GalileanSpatialTranslationOperator", "GalileanTranslationOperator"]


from dataclasses import replace
from typing import Any, Literal, final
from typing_extensions import override

import equinox as eqx
from jaxtyping import ArrayLike
from plum import convert, dispatch

import quaxed.numpy as jnp
from unxt import Quantity

from .base import AbstractGalileanOperator
from coordinax._src.base import AbstractPosition
from coordinax._src.d1.cartesian import CartesianPosition1D
from coordinax._src.d2.cartesian import CartesianPosition2D
from coordinax._src.d3.base import AbstractPosition3D
from coordinax._src.d3.cartesian import CartesianPosition3D
from coordinax._src.d4.spacetime import FourVector
from coordinax._src.operators.base import AbstractOperator, op_call_dispatch
from coordinax._src.operators.funcs import simplify_op
from coordinax._src.operators.identity import IdentityOperator

##############################################################################
# Spatial Translations


def _converter_spatialtranslation(x: Any) -> AbstractPosition:
    """Convert to a spatial translation vector."""
    out: AbstractPosition | None = None
    if isinstance(x, GalileanSpatialTranslationOperator):
        out = x.translation
    elif isinstance(x, AbstractPosition):
        out = x
    elif isinstance(x, Quantity):
        shape: tuple[int, ...] = x.shape
        match shape:
            case (1,):
                out = CartesianPosition1D.constructor(x)
            case (2,):
                out = CartesianPosition2D.constructor(x)
            case (3,):
                out = CartesianPosition3D.constructor(x)
            case _:
                msg = f"Cannot convert {x} to a spatial translation vector."
                raise TypeError(msg)

    if out is None:
        msg = f"Cannot convert {x} to a spatial translation vector."
        raise TypeError(msg)

    return out


@final
class GalileanSpatialTranslationOperator(AbstractGalileanOperator):
    r"""Operator for Galilean spatial translations.

    The coordinate transform is given by:

    .. math::

        (t,\mathbf{x}) \mapsto (t+s, \mathbf{x} + \mathbf {a})

    Parameters
    ----------
    translation : :class:`vector.AbstractPosition3D`
        The spatial translation vector. This parameters accepts either a
        :class:`vector.AbstractPosition3D` instance or uses
        :meth:`vector.CartesianPosition3D.constructor` to enable a variety of more
        convenient input types to create a Cartesian vector. See
        :class:`vector.CartesianPosition3D` for details.

    Examples
    --------
    We start with the required imports:

    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity
    >>> import coordinax as cx

    We can then create a spatial translation operator:

    >>> shift = Quantity([1.0, 2.0, 3.0], "kpc")
    >>> op = cx.operators.GalileanSpatialTranslationOperator(shift)
    >>> op
    GalileanSpatialTranslationOperator( translation=CartesianPosition3D( ... ) )

    Note that the translation is a :class:`vector.CartesianPosition3D`, which was
    constructed from a 1D array, using
    :meth:`vector.CartesianPosition3D.constructor`. We can also construct it
    directly, which allows for other vector types.

    >>> shift = cx.SphericalPosition(r=Quantity(1.0, "kpc"),
    ...                              theta=Quantity(jnp.pi/2, "rad"),
    ...                              phi=Quantity(0, "rad"))
    >>> op = cx.operators.GalileanSpatialTranslationOperator(shift)
    >>> op
    GalileanSpatialTranslationOperator( translation=SphericalPosition( ... ) )

    Translation operators can be applied to :class:`vector.AbstractPosition`:

    >>> q = cx.CartesianPosition3D.constructor([0, 0, 0], "kpc")
    >>> op(q)
    CartesianPosition3D( ... )

    And to :class:`~unxt.Quantity`:

    >>> q = Quantity([0, 0, 0], "kpc")
    >>> op(q).value.round(2)
    Array([ 1.,  0., -0.], dtype=float32)

    :class:`coordinax.operators.GalileanSpatialTranslationOperator` can be used
    for other dimensional vectors as well:

    - 1D:

    >>> shift = Quantity([1], "kpc")
    >>> op = cx.operators.GalileanSpatialTranslationOperator(shift)
    >>> q = Quantity([0], "kpc")
    >>> op(q)
    Quantity['length'](Array([1.], dtype=float32), unit='kpc')

    >>> vec = cx.CartesianPosition1D.constructor(q).represent_as(cx.RadialPosition)
    >>> op(vec)
    RadialPosition(r=Distance(value=f32[], unit=Unit("kpc")))

    - 2D:

    >>> shift = Quantity([1, 2], "kpc")
    >>> op = cx.operators.GalileanSpatialTranslationOperator(shift)
    >>> q = Quantity([0, 0], "kpc")
    >>> op(q)
    Quantity['length'](Array([1., 2.], dtype=float32), unit='kpc')

    >>> vec = cx.CartesianPosition2D.constructor(q).represent_as(cx.PolarPosition)
    >>> op(vec)
    PolarPosition( r=Distance(value=f32[], unit=Unit("kpc")),
                 phi=Quantity[...](value=f32[], unit=Unit("rad")) )

    - 3D:

    >>> shift = Quantity([1, 2, 3], "kpc")
    >>> op = cx.operators.GalileanSpatialTranslationOperator(shift)
    >>> q = Quantity([0, 0, 0], "kpc")
    >>> op(q)
    Quantity['length'](Array([1., 2., 3.], dtype=float32), unit='kpc')

    >>> vec = cx.CartesianPosition3D.constructor(q).represent_as(cx.SphericalPosition)
    >>> op(vec)
    SphericalPosition( r=Distance(value=f32[], unit=Unit("kpc")),
                     theta=Quantity[...](value=f32[], unit=Unit("rad")),
                     phi=Quantity[...](value=f32[], unit=Unit("rad")) )

    Many operators are time dependent and require a time argument. This operator
    is time independent and will pass through the time argument:

    >>> t = Quantity(0, "Gyr")
    >>> op(q, t)[1] is t
    True

    """

    translation: AbstractPosition = eqx.field(converter=_converter_spatialtranslation)
    """The spatial translation.

    This parameters accepts either a :class:`vector.AbstracVector` instance or
    uses a Cartesian vector constructor to enable a variety of more convenient
    input types to create a Cartesian vector. See
    :class:`vector.CartesianPosition3D.constructor` for an example when doing a 3D
    translation.
    """

    # -------------------------------------------

    @override
    @classmethod
    @dispatch
    def constructor(
        cls: "type[GalileanSpatialTranslationOperator]",
        x: ArrayLike | list[float | int],
        unit: str,  # TODO: support unit object
        /,
    ) -> "GalileanSpatialTranslationOperator":
        """Construct a spatial translation operator.

        Examples
        --------
        >>> from unxt import Quantity
        >>> import coordinax as cx

        >>> op = cx.operators.GalileanSpatialTranslationOperator.constructor([1, 1, 1], "kpc")
        >>> op.translation.x
        Quantity['length'](Array(1., dtype=float32), unit='kpc')

        """  # noqa: E501
        return cls(Quantity(x, unit))

    @override
    @classmethod
    @dispatch
    def constructor(
        cls: "type[GalileanSpatialTranslationOperator]",
        x: ArrayLike | list[float | int],
        *,
        unit: Any,
    ) -> "GalileanSpatialTranslationOperator":
        """Construct a spatial translation operator.

        Examples
        --------
        >>> from unxt import Quantity
        >>> import coordinax as cx

        >>> op = cx.operators.GalileanSpatialTranslationOperator.constructor([1, 1, 1], "kpc")
        >>> op.translation.x
        Quantity['length'](Array(1., dtype=float32), unit='kpc')

        """  # noqa: E501
        return cls(Quantity(x, unit))

    # -------------------------------------------

    @property
    def is_inertial(self) -> Literal[True]:
        """Galilean translation is an inertial frame-preserving transformation.

        Examples
        --------
        >>> from unxt import Quantity
        >>> import coordinax as cx
        >>> from coordinax.operators import GalileanSpatialTranslationOperator

        >>> shift = cx.CartesianPosition3D.constructor([1, 1, 1], "kpc")
        >>> op = GalileanSpatialTranslationOperator(shift)

        >>> op.is_inertial
        True

        """
        return True

    @property
    def inverse(self) -> "GalileanSpatialTranslationOperator":
        """The inverse of the operator.

        Examples
        --------
        >>> from unxt import Quantity
        >>> import coordinax as cx
        >>> from coordinax.operators import GalileanSpatialTranslationOperator

        >>> shift = cx.CartesianPosition3D.constructor([1, 1, 1], "kpc")
        >>> op = GalileanSpatialTranslationOperator(shift)

        >>> op.inverse
        GalileanSpatialTranslationOperator( translation=CartesianPosition3D( ... ) )

        >>> op.inverse.translation.x
        Quantity['length'](Array(-1., dtype=float32), unit='kpc')

        """
        return GalileanSpatialTranslationOperator(-self.translation)

    # -------------------------------------------

    @op_call_dispatch(precedence=1)
    def __call__(
        self: "GalileanSpatialTranslationOperator", q: AbstractPosition, /
    ) -> AbstractPosition:
        """Apply the translation to the coordinates.

        Examples
        --------
        >>> from unxt import Quantity
        >>> import coordinax as cx
        >>> from coordinax.operators import GalileanSpatialTranslationOperator

        >>> shift = cx.CartesianPosition3D.constructor([1, 1, 1], "kpc")
        >>> op = GalileanSpatialTranslationOperator(shift)

        >>> q = cx.CartesianPosition3D.constructor([1, 2, 3], "kpc")
        >>> t = Quantity(0, "Gyr")
        >>> newq = op(q)
        >>> newq.x
        Quantity['length'](Array(2., dtype=float32), unit='kpc')

        """
        return q + self.translation

    @op_call_dispatch(precedence=1)
    def __call__(
        self: "GalileanSpatialTranslationOperator",
        q: AbstractPosition,
        t: Quantity["time"],
        /,
    ) -> tuple[AbstractPosition, Quantity["time"]]:
        """Apply the translation to the coordinates.

        Examples
        --------
        >>> from unxt import Quantity
        >>> import coordinax as cx
        >>> from coordinax.operators import GalileanSpatialTranslationOperator

        >>> shift = cx.CartesianPosition3D.constructor([1, 1, 1], "kpc")
        >>> op = GalileanSpatialTranslationOperator(shift)

        >>> q = cx.CartesianPosition3D.constructor([1, 2, 3], "kpc")
        >>> t = Quantity(0, "Gyr")
        >>> newq, newt = op(q, t)
        >>> newq.x
        Quantity['length'](Array(2., dtype=float32), unit='kpc')

        This spatial translation is time independent.

        >>> op(q, Quantity(1, "Gyr"))[0].x == newq.x
        Array(True, dtype=bool)

        """
        return q + self.translation, t

    @op_call_dispatch(precedence=1)
    def __call__(
        self: "GalileanSpatialTranslationOperator", v4: FourVector, /
    ) -> AbstractPosition:
        """Apply the translation to the coordinates."""  # TODO: docstring
        return replace(v4, q=v4.q + self.translation)


@simplify_op.register
def _simplify_op_spatialtranslation(
    op: GalileanSpatialTranslationOperator, /, **kwargs: Any
) -> AbstractOperator:
    """Simplify a spatial translation operator."""
    # Check if the translation is zero.
    if jnp.allclose(convert(op.translation, Quantity).value, jnp.zeros((3,)), **kwargs):
        return IdentityOperator()
    return op


##############################################################################


@final
class GalileanTranslationOperator(AbstractGalileanOperator):
    r"""Operator for spatio-temporal translations.

    The coordinate transform is given by:

    .. math::

        (t,\mathbf{x}) \mapsto (t+s, \mathbf{x} + \mathbf {a})

    where :math:`a \in R^3` and :math:`s \in R`.  Therefore for a potential
    :math:`\Phi(t,\mathbf{x})` in the translated frame the potential is given by
    the subtraction of the translation.

    Parameters
    ----------
    translation : :class:`vector.FourVector`
        The translation vector [T, Q].  This parameters uses
        :meth:`vector.FourVector.constructor` to enable a variety of more
        convenient input types. See :class:`vector.FourVector` for details.

    Examples
    --------
    We start with the required imports:

    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity
    >>> import coordinax as cx
    >>> import coordinax.operators as co

    We can then create a translation operator:

    >>> op = GalileanTranslationOperator(Quantity([1.0, 2.0, 3.0, 4.0], "kpc"))
    >>> op
    GalileanTranslationOperator(
      translation=FourVector(
        t=Quantity[PhysicalType('time')](value=f32[], unit=Unit("kpc s / km")),
        q=CartesianPosition3D( ... ) )
    )

    Note that the translation is a :class:`vector.FourVector`, which was
    constructed from a 1D array, using :meth:`vector.FourVector.constructor`. We
    can also construct it directly, which allows for other vector types.

    >>> qshift = cx.SphericalPosition(r=Quantity(1.0, "kpc"),
    ...                               theta=Quantity(jnp.pi/2, "rad"),
    ...                               phi=Quantity(0, "rad"))
    >>> op = GalileanTranslationOperator(FourVector(t=Quantity(1.0, "Gyr"), q=qshift))
    >>> op
    GalileanTranslationOperator(
      translation=FourVector(
        t=Quantity[PhysicalType('time')](value=f32[], unit=Unit("Gyr")),
        q=SphericalPosition( ... ) )
    )

    Translation operators can be applied to :class:`vector.FourVector`:

    >>> w = FourVector.constructor([0, 0, 0, 0], "kpc")
    >>> op(w)
    FourVector(
      t=Quantity[PhysicalType('time')](value=f32[], unit=Unit("kpc s / km")),
      q=CartesianPosition3D( ... )
    )

    Also to :class:`vector.AbstractPosition3D` and :class:`unxt.Quantity`:

    >>> q = cx.CartesianPosition3D.constructor([0, 0, 0], "kpc")
    >>> t = Quantity(0, "Gyr")
    >>> newq, newt = op(q, t)
    >>> newq.x
    Quantity['length'](Array(1., dtype=float32), unit='kpc')
    >>> newt
    Quantity['time'](Array(1., dtype=float32), unit='Gyr')

    """

    translation: FourVector = eqx.field(converter=FourVector.constructor)
    """The temporal + spatial translation.

    The translation vector [T, Q].  This parameters uses
    :meth:`vector.FourVector.constructor` to enable a variety of more convenient
    input types. See :class:`vector.FourVector` for details.
    """

    # -------------------------------------------

    @property
    def is_inertial(self) -> Literal[True]:
        """Galilean translation is an inertial-frame preserving transformation.

        Examples
        --------
        >>> from unxt import Quantity
        >>> import coordinax as cx
        >>> from coordinax.operators import GalileanTranslationOperator

        >>> shift = cx.FourVector.constructor([0, 1, 1, 1], "kpc")
        >>> op = GalileanTranslationOperator(shift)

        >>> op.is_inertial
        True

        """
        return True

    @property
    def inverse(self) -> "GalileanTranslationOperator":
        """The inverse of the operator.

        Examples
        --------
        >>> from unxt import Quantity
        >>> import coordinax as cx
        >>> from coordinax.operators import GalileanSpatialTranslationOperator

        >>> qshift = cx.CartesianPosition3D.constructor([1, 1, 1], "kpc")
        >>> tshift = Quantity(1, "Gyr")
        >>> shift = FourVector(tshift, qshift)
        >>> op = GalileanTranslationOperator(shift)

        >>> op.inverse
        GalileanTranslationOperator( translation=FourVector( ... ) )

        >>> op.inverse.translation.q.x
        Quantity['length'](Array(-1., dtype=float32), unit='kpc')

        """
        return GalileanTranslationOperator(-self.translation)

    # -------------------------------------------

    @op_call_dispatch
    def __call__(self: "GalileanTranslationOperator", x: FourVector, /) -> FourVector:
        """Apply the translation to the coordinates.

        Examples
        --------
        >>> from unxt import Quantity
        >>> import coordinax as cx
        >>> from coordinax.operators import GalileanTranslationOperator

        Explicitly construct the translation operator:

        >>> qshift = cx.CartesianPosition3D.constructor([1, 1, 1], "kpc")
        >>> tshift = Quantity(1, "Gyr")
        >>> shift = FourVector(tshift, qshift)
        >>> op = GalileanTranslationOperator(shift)

        Construct a vector to translate, using the convenience constructor (the
        0th component is :math:`c * t`, the rest are spatial components):

        >>> w = cx.FourVector.constructor([0, 1, 2, 3], "kpc")
        >>> w.t
        Quantity['time'](Array(0., dtype=float32), unit='kpc s / km')

        Apply the translation operator:

        >>> new = op(w)
        >>> new.x
        Quantity['length'](Array(2., dtype=float32), unit='kpc')

        >>> new.t.to_units("Gyr")
        Quantity['time'](Array(0.99999994, dtype=float32), unit='Gyr')

        """
        return x + self.translation

    @op_call_dispatch(precedence=1)
    def __call__(
        self: "GalileanTranslationOperator",
        x: AbstractPosition3D,
        t: Quantity["time"],
        /,
    ) -> tuple[AbstractPosition3D, Quantity["time"]]:
        """Apply the translation to the coordinates.

        Examples
        --------
        >>> from unxt import Quantity
        >>> import coordinax as cx
        >>> from coordinax.operators import GalileanTranslationOperator

        Explicitly construct the translation operator:

        >>> qshift = cx.CartesianPosition3D.constructor([1, 1, 1], "kpc")
        >>> tshift = Quantity(1, "Gyr")
        >>> shift = cx.FourVector(tshift, qshift)
        >>> op = GalileanTranslationOperator(shift)

        Construct a vector to translate

        >>> q = cx.CartesianPosition3D.constructor([1, 2, 3], "kpc")
        >>> t = Quantity(1, "Gyr")
        >>> newq, newt = op(q, t)

        >>> newq.x
        Quantity['length'](Array(2., dtype=float32), unit='kpc')

        >>> newt
        Quantity['time'](Array(2., dtype=float32), unit='Gyr')

        """
        return (x + self.translation.q, t + self.translation.t)


@simplify_op.register
def _simplify_op_tranlation(
    op: GalileanTranslationOperator, /, **kwargs: Any
) -> AbstractOperator:
    """Simplify a translation operator."""
    # Check if the translation is zero.
    if jnp.allclose(convert(op.translation, Quantity).value, jnp.zeros((4,)), **kwargs):
        return IdentityOperator()
    # Check if the translation is purely spatial.
    if op.translation[0] == 0:
        return GalileanSpatialTranslationOperator(op.translation[1:])
    return op



"""Test :mod:`coordinax.utils`."""

from warnings import warn

import pytest

from coordinax._src.exceptions import IrreversibleDimensionChange


def test_warning_irreversibledimensionchange():
    """Test :class:`IrreversibleDimensionChange`."""
    with pytest.warns(UserWarning, match="Irreversible dimension change"):
        warn("Irreversible dimension change", IrreversibleDimensionChange, stacklevel=2)



"""Representation of coordinates in different systems."""

__all__: list[str] = []

from typing import Any

from plum import dispatch

from .cartesian import CartesianAccelerationND, CartesianPositionND, CartesianVelocityND
from .poincare import PoincarePolarVector

###############################################################################
# Cartesian


@dispatch
def represent_as(
    current: CartesianPositionND, target: type[CartesianPositionND], /, **kwargs: Any
) -> CartesianPositionND:
    """CartesianPositionND -> CartesianPositionND.

    Examples
    --------
    >>> import coordinax as cx
    >>> from unxt import Quantity

    >>> q = cx.CartesianPositionND(Quantity([1, 2, 3, 4], "kpc"))

    >>> cx.represent_as(q, cx.CartesianPositionND) is q
    True

    """
    return current


@dispatch
def represent_as(
    current: CartesianVelocityND,
    target: type[CartesianVelocityND],
    position: CartesianPositionND,
    /,
    **kwargs: Any,
) -> CartesianVelocityND:
    """CartesianVelocityND -> CartesianVelocityND.

    Examples
    --------
    >>> import coordinax as cx
    >>> from unxt import Quantity

    >>> x = cx.CartesianPositionND(Quantity([1, 2, 3, 4], "km"))
    >>> v = cx.CartesianVelocityND(Quantity([1, 2, 3, 4], "km/s"))

    >>> cx.represent_as(v, cx.CartesianVelocityND, x) is v
    True

    """
    return current


# =============================================================================
# CartesianVelocityND


@dispatch
def represent_as(
    current: CartesianVelocityND, target: type[CartesianVelocityND], /
) -> CartesianVelocityND:
    """CartesianVelocityND -> CartesianVelocityND with no position.

    Cartesian coordinates are an affine coordinate system and so the
    transformation of an n-th order derivative vector in this system do not
    require lower-order derivatives to be specified. See
    https://en.wikipedia.org/wiki/Tensors_in_curvilinear_coordinates for more
    information. This mixin provides a corresponding implementation of the
    `coordinax.represent_as` method for Cartesian velocities.

    Examples
    --------
    >>> import coordinax as cx
    >>> v = cx.CartesianVelocityND.constructor([1, 1, 1], "m/s")
    >>> cx.represent_as(v, cx.CartesianVelocityND) is v
    True

    """
    return current


# =============================================================================
# CartesianAccelerationND


@dispatch
def represent_as(
    current: CartesianAccelerationND, target: type[CartesianAccelerationND], /
) -> CartesianAccelerationND:
    """CartesianAccelerationND -> CartesianAccelerationND with no position.

    Cartesian coordinates are an affine coordinate system and so the
    transformation of an n-th order derivative vector in this system do not
    require lower-order derivatives to be specified. See
    https://en.wikipedia.org/wiki/Tensors_in_curvilinear_coordinates for more
    information. This mixin provides a corresponding implementation of the
    `coordinax.represent_as` method for Cartesian vectors.

    Examples
    --------
    >>> import coordinax as cx
    >>> a = cx.CartesianAccelerationND.constructor([1, 1, 1], "m/s2")
    >>> cx.represent_as(a, cx.CartesianAccelerationND) is a
    True

    """
    return current


###############################################################################
# Poincare


@dispatch
def represent_as(
    current: PoincarePolarVector, target: type[PoincarePolarVector], /, **kwargs: Any
) -> PoincarePolarVector:
    """PoincarePolarVector -> PoincarePolarVector."""
    return current



"""Built-in vector classes."""

__all__ = [
    "CylindricalPosition",
    "CylindricalVelocity",
    "CylindricalAcceleration",
]

from functools import partial
from typing import final
from typing_extensions import override

import equinox as eqx

import quaxed.numpy as xp
from unxt import Quantity

import coordinax._src.typing as ct
from .base import AbstractAcceleration3D, AbstractPosition3D, AbstractVelocity3D
from coordinax._src.checks import check_azimuth_range, check_r_non_negative
from coordinax._src.converters import converter_azimuth_to_range
from coordinax._src.utils import classproperty


@final
class CylindricalPosition(AbstractPosition3D):
    """Cylindrical vector representation.

    This adheres to ISO standard 31-11.

    """

    rho: ct.BatchableLength = eqx.field(
        converter=partial(Quantity["length"].constructor, dtype=float)
    )
    r"""Cylindrical radial distance :math:`\rho \in [0,+\infty)`."""

    phi: ct.BatchableAngle = eqx.field(
        converter=lambda x: converter_azimuth_to_range(
            Quantity["angle"].constructor(x, dtype=float)  # pylint: disable=E1120
        )
    )
    r"""Azimuthal angle :math:`\phi \in [0,360)`."""

    z: ct.BatchableLength = eqx.field(
        converter=partial(Quantity["length"].constructor, dtype=float)
    )
    r"""Height :math:`z \in (-\infty,+\infty)`."""

    def __check_init__(self) -> None:
        """Check the validity of the initialisation."""
        check_r_non_negative(self.rho)
        check_azimuth_range(self.phi)

    @override
    @classproperty
    @classmethod
    def differential_cls(cls) -> type["CylindricalVelocity"]:
        return CylindricalVelocity

    @override
    @partial(eqx.filter_jit, inline=True)
    def norm(self) -> ct.BatchableLength:
        """Return the norm of the vector.

        Examples
        --------
        >>> from unxt import Quantity
        >>> import coordinax as cx
        >>> c = cx.CylindricalPosition(rho=Quantity(3, "kpc"), phi=Quantity(0, "deg"),
        ...                       z=Quantity(4, "kpc"))
        >>> c.norm()
        Quantity['length'](Array(5., dtype=float32), unit='kpc')

        """
        return xp.sqrt(self.rho**2 + self.z**2)


@final
class CylindricalVelocity(AbstractVelocity3D):
    """Cylindrical differential representation."""

    d_rho: ct.BatchableSpeed = eqx.field(
        converter=partial(Quantity["speed"].constructor, dtype=float)
    )
    r"""Cyindrical radial speed :math:`d\rho/dt \in [-\infty, \infty]."""

    d_phi: ct.BatchableAngularSpeed = eqx.field(
        converter=partial(Quantity["angular speed"].constructor, dtype=float)
    )
    r"""Azimuthal speed :math:`d\phi/dt \in [-\infty, \infty]."""

    d_z: ct.BatchableSpeed = eqx.field(
        converter=partial(Quantity["speed"].constructor, dtype=float)
    )
    r"""Vertical speed :math:`dz/dt \in [-\infty, \infty]."""

    @classproperty
    @classmethod
    def integral_cls(cls) -> type[CylindricalPosition]:
        return CylindricalPosition

    @classproperty
    @classmethod
    def differential_cls(cls) -> type["CylindricalAcceleration"]:
        return CylindricalAcceleration


@final
class CylindricalAcceleration(AbstractAcceleration3D):
    """Cylindrical acceleration representation."""

    d2_rho: ct.BatchableSpeed = eqx.field(
        converter=partial(Quantity["acceleration"].constructor, dtype=float)
    )
    r"""Cyindrical radial acceleration :math:`d^2\rho/dt^2 \in [-\infty, \infty]."""

    d2_phi: ct.BatchableAngularSpeed = eqx.field(
        converter=partial(Quantity["angular acceleration"].constructor, dtype=float)
    )
    r"""Azimuthal acceleration :math:`d^2\phi/dt^2 \in [-\infty, \infty]."""

    d2_z: ct.BatchableSpeed = eqx.field(
        converter=partial(Quantity["acceleration"].constructor, dtype=float)
    )
    r"""Vertical acceleration :math:`d^2z/dt^2 \in [-\infty, \infty]."""

    @classproperty
    @classmethod
    def integral_cls(cls) -> type[CylindricalVelocity]:
        return CylindricalVelocity



"""Intra-ecosystem Compatibility."""

__all__: list[str] = []


from jaxtyping import Shaped
from plum import conversion_method, convert

import quaxed.numpy as jnp
from unxt import Quantity

from .spacetime import FourVector
from coordinax._src.operators.base import AbstractOperator, op_call_dispatch

#####################################################################
# Convert to Quantity


@conversion_method(type_from=FourVector, type_to=Quantity)  # type: ignore[misc]
def vec_to_q(obj: FourVector, /) -> Shaped[Quantity["length"], "*batch 4"]:
    """`coordinax.AbstractPosition3D` -> `unxt.Quantity`."""
    cart = convert(obj.q, Quantity)
    return jnp.concat([obj.c * obj.t[..., None], cart], axis=-1)


#####################################################################
# Operators


@op_call_dispatch
def call(self: AbstractOperator, v4: FourVector, /) -> FourVector:
    """Dispatch to the operator's `__call__` method.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    We can then create a spatial translation operator:

    >>> op = cx.operators.GalileanSpatialTranslationOperator(Quantity([1, 2, 3], "kpc"))
    >>> op
    GalileanSpatialTranslationOperator( translation=CartesianPosition3D( ... ) )

    We can then apply the operator to a position:

    >>> pos = cx.FourVector.constructor([0, 1.0, 2.0, 3.0], "kpc")
    >>> pos
    FourVector( t=Quantity[PhysicalType('time')](...), q=CartesianPosition3D( ... ) )

    >>> newpos = op(pos)
    >>> newpos
    FourVector( t=Quantity[PhysicalType('time')](...), q=CartesianPosition3D( ... ) )
    >>> newpos.q.x
    Quantity['length'](Array(2., dtype=float32), unit='kpc')

    """
    q, t = self(v4.q, v4.t)
    return FourVector(t=t, q=q)


@op_call_dispatch
def call(
    self: AbstractOperator, x: Shaped[Quantity["length"], "*batch 4"], /
) -> Shaped[Quantity["length"], "*batch 4"]:
    """Dispatch to the operator's `__call__` method.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    We can then create a spatial translation operator:

    >>> op = cx.operators.GalileanSpatialTranslationOperator(Quantity([1, 2, 3], "kpc"))
    >>> op
    GalileanSpatialTranslationOperator( translation=CartesianPosition3D( ... ) )

    We can then apply the operator to a position:

    >>> pos = Quantity([0, 1.0, 2.0, 3.0], "kpc")
    >>> pos
    Quantity['length'](Array([0., 1., 2., 3.], dtype=float32), unit='kpc')

    >>> newpos = op(pos)
    >>> newpos
    Quantity['length'](Array([0., 2., 4., 6.], dtype=float32), unit='kpc')

    """
    return convert(self(FourVector.constructor(x)), Quantity)



"""Built-in vector classes."""

__all__ = [
    "CartesianGeneric3D",
]

from dataclasses import fields
from functools import partial
from typing import Any, TypeVar, final

import equinox as eqx
import jax

import quaxed.numpy as jnp
from unxt import AbstractQuantity, Quantity

import coordinax._src.typing as ct
from coordinax._src.base import AbstractVector
from coordinax._src.base.mixins import AvalMixin

VT = TypeVar("VT", bound="CartesianGeneric3D")


@final
class CartesianGeneric3D(AvalMixin, AbstractVector):
    """Generic 3D Cartesian coordinates.

    The fields of this class are not restricted to any specific dimensions.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> vec = cx.CartesianGeneric3D.constructor([1, 2, 3], "kg m /s")
    >>> vec
    CartesianGeneric3D(
      x=Quantity[...]( value=f32[], unit=Unit("kg m / s") ),
      y=Quantity[...]( value=f32[], unit=Unit("kg m / s") ),
      z=Quantity[...]( value=f32[], unit=Unit("kg m / s") )
    )

    """

    x: ct.BatchableFloatScalarQ = eqx.field(
        converter=partial(Quantity.constructor, dtype=float)
    )

    y: ct.BatchableFloatScalarQ = eqx.field(
        converter=partial(Quantity.constructor, dtype=float)
    )

    z: ct.BatchableFloatScalarQ = eqx.field(
        converter=partial(Quantity.constructor, dtype=float)
    )

    # -----------------------------------------------------
    # Unary operations

    def __neg__(self) -> "CartesianGeneric3D":
        """Negate the `coordinax.CartesianGeneric3D`.

        Examples
        --------
        >>> import coordinax as cx
        >>> q = cx.CartesianGeneric3D.constructor([1, 2, 3], "kpc")
        >>> (-q).x
        Quantity['length'](Array(-1., dtype=float32), unit='kpc')

        """
        return jax.tree.map(jnp.negative, self)

    # ===============================================================
    # Convenience methods

    def represent_as(self, target: type[VT], /, *args: Any, **kwargs: Any) -> VT:
        """Represent the vector as another type."""
        raise NotImplementedError  # pragma: no cover


# =====================================================
# Constructors


@CartesianGeneric3D.constructor._f.dispatch  # type: ignore[attr-defined, misc]  # noqa: SLF001
def constructor(
    cls: type[CartesianGeneric3D],
    obj: AbstractQuantity,  # TODO: Shaped[AbstractQuantity, "*batch 3"]
    /,
) -> CartesianGeneric3D:
    """Construct a 3D Cartesian position.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> vec = cx.CartesianGeneric3D.constructor(Quantity([1, 2, 3], "m"))
    >>> vec
    CartesianGeneric3D(
      x=Quantity[...](value=f32[], unit=Unit("m")),
      y=Quantity[...](value=f32[], unit=Unit("m")),
      z=Quantity[...](value=f32[], unit=Unit("m"))
    )

    """
    comps = {f.name: obj[..., i] for i, f in enumerate(fields(cls))}
    return cls(**comps)



"""Built-in vector classes."""

__all__ = [
    "CartesianPosition2D",
    "CartesianVelocity2D",
    "CartesianAcceleration2D",
]

from dataclasses import fields, replace
from functools import partial
from typing import final
from typing_extensions import override

import equinox as eqx
import jax
from jaxtyping import ArrayLike, Shaped
from quax import register

import quaxed.numpy as jnp
from quaxed import lax as qlax
from unxt import AbstractQuantity, Quantity

import coordinax._src.typing as ct
from .base import AbstractAcceleration2D, AbstractPosition2D, AbstractVelocity2D
from coordinax._src.base import AbstractPosition
from coordinax._src.base.mixins import AvalMixin
from coordinax._src.utils import classproperty


@final
class CartesianPosition2D(AbstractPosition2D):
    """Cartesian vector representation."""

    x: ct.BatchableLength = eqx.field(
        converter=partial(Quantity["length"].constructor, dtype=float)
    )
    r"""X coordinate :math:`x \in (-\infty,+\infty)`."""

    y: ct.BatchableLength = eqx.field(
        converter=partial(Quantity["length"].constructor, dtype=float)
    )
    r"""Y coordinate :math:`y \in (-\infty,+\infty)`."""

    @classproperty
    @classmethod
    def differential_cls(cls) -> type["CartesianVelocity2D"]:
        return CartesianVelocity2D


# -----------------------------------------------------


@CartesianPosition2D.constructor._f.dispatch  # type: ignore[attr-defined, misc] # noqa: SLF001
def constructor(
    cls: type[CartesianPosition2D], obj: Shaped[AbstractQuantity, "*batch 2"], /
) -> CartesianPosition2D:
    """Construct a 2D Cartesian position.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> vec = cx.CartesianPosition2D.constructor(Quantity([1, 2], "m"))
    >>> vec
    CartesianPosition2D(
        x=Quantity[...](value=f32[], unit=Unit("m")),
        y=Quantity[...](value=f32[], unit=Unit("m"))
    )

    """
    comps = {f.name: obj[..., i] for i, f in enumerate(fields(cls))}
    return cls(**comps)


# -----------------------------------------------------


@register(jax.lax.add_p)  # type: ignore[misc]
def _add_cart2d_pos(
    lhs: CartesianPosition2D, rhs: AbstractPosition, /
) -> CartesianPosition2D:
    """Add two vectors.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> cart = cx.CartesianPosition2D.constructor(Quantity([1, 2], "kpc"))
    >>> polr = cx.PolarPosition(r=Quantity(3, "kpc"), phi=Quantity(90, "deg"))
    >>> (cart + polr).x
    Quantity['length'](Array(0.9999999, dtype=float32), unit='kpc')

    >>> jnp.add(cart, polr).x
    Quantity['length'](Array(0.9999999, dtype=float32), unit='kpc')

    """
    cart = rhs.represent_as(CartesianPosition2D)
    return jax.tree.map(qlax.add, lhs, cart)


@register(jax.lax.mul_p)  # type: ignore[misc]
def _mul_v_cart2d(lhs: ArrayLike, rhs: CartesianPosition2D, /) -> CartesianPosition2D:
    """Scale a cartesian 2D position by a scalar.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> v = cx.CartesianPosition2D.constructor(Quantity([3, 4], "m"))
    >>> jnp.multiply(5, v).x
    Quantity['length'](Array(15., dtype=float32), unit='m')

    """
    # Validation
    lhs = eqx.error_if(
        lhs, any(jax.numpy.shape(lhs)), f"must be a scalar, not {type(lhs)}"
    )

    # Scale the components
    return replace(rhs, x=lhs * rhs.x, y=lhs * rhs.y)


@register(jax.lax.neg_p)  # type: ignore[misc]
def _neg_p_cart2d_pos(obj: CartesianPosition2D, /) -> CartesianPosition2D:
    """Negate the `coordinax.CartesianPosition2D`.

    Examples
    --------
    >>> import coordinax as cx
    >>> q = cx.CartesianPosition2D.constructor([1, 2], "km")
    >>> (-q).x
    Quantity['length'](Array(-1., dtype=float32), unit='km')

    """
    return jax.tree.map(qlax.neg, obj)


@register(jax.lax.sub_p)  # type: ignore[misc]
def _sub_cart2d_pos2d(
    lhs: CartesianPosition2D, rhs: AbstractPosition, /
) -> CartesianPosition2D:
    """Subtract two vectors.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx
    >>> cart = cx.CartesianPosition2D.constructor([1, 2], "kpc")
    >>> polr = cx.PolarPosition(r=Quantity(3, "kpc"), phi=Quantity(90, "deg"))

    >>> (cart - polr).x
    Quantity['length'](Array(1.0000001, dtype=float32), unit='kpc')

    """
    cart = rhs.represent_as(CartesianPosition2D)
    return jax.tree.map(qlax.sub, lhs, cart)


#####################################################################


@final
class CartesianVelocity2D(AvalMixin, AbstractVelocity2D):
    """Cartesian differential representation."""

    d_x: ct.BatchableSpeed = eqx.field(
        converter=partial(Quantity["speed"].constructor, dtype=float)
    )
    r"""X coordinate differential :math:`\dot{x} \in (-\infty,+\infty)`."""

    d_y: ct.BatchableSpeed = eqx.field(
        converter=partial(Quantity["speed"].constructor, dtype=float)
    )
    r"""Y coordinate differential :math:`\dot{y} \in (-\infty,+\infty)`."""

    @classproperty
    @classmethod
    def integral_cls(cls) -> type[CartesianPosition2D]:
        return CartesianPosition2D

    @classproperty
    @classmethod
    def differential_cls(cls) -> type["CartesianAcceleration2D"]:
        return CartesianAcceleration2D


# -----------------------------------------------------


@CartesianVelocity2D.constructor._f.dispatch  # type: ignore[attr-defined, misc] # noqa: SLF001
def constructor(
    cls: type[CartesianVelocity2D], obj: Shaped[AbstractQuantity, "*batch 2"], /
) -> CartesianVelocity2D:
    """Construct a 2D Cartesian velocity.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> vec = cx.CartesianVelocity2D.constructor(Quantity([1, 2], "m/s"))
    >>> vec
    CartesianVelocity2D(
      d_x=Quantity[...]( value=f32[], unit=Unit("m / s") ),
      d_y=Quantity[...]( value=f32[], unit=Unit("m / s") )
    )

    """
    comps = {f.name: obj[..., i] for i, f in enumerate(fields(cls))}
    return cls(**comps)


# -----------------------------------------------------


@register(jax.lax.add_p)  # type: ignore[misc]
def _add_pp(
    lhs: CartesianVelocity2D, rhs: CartesianVelocity2D, /
) -> CartesianVelocity2D:
    """Add two Cartesian velocities.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> v = cx.CartesianVelocity2D.constructor(Quantity([1, 2], "km/s"))
    >>> (v + v).d_x
    Quantity['speed'](Array(2., dtype=float32), unit='km / s')

    >>> jnp.add(v, v).d_x
    Quantity['speed'](Array(2., dtype=float32), unit='km / s')

    """
    return jax.tree.map(qlax.add, lhs, rhs)


@register(jax.lax.mul_p)  # type: ignore[misc]
def _mul_vp(lhs: ArrayLike, rhts: CartesianVelocity2D, /) -> CartesianVelocity2D:
    """Scale a cartesian 2D velocity by a scalar.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> v = cx.CartesianVelocity2D.constructor(Quantity([3, 4], "m/s"))
    >>> (5 * v).d_x
    Quantity['speed'](Array(15., dtype=float32), unit='m / s')

    >>> jnp.multiply(5, v).d_x
    Quantity['speed'](Array(15., dtype=float32), unit='m / s')

    """
    # Validation
    lhs = eqx.error_if(
        lhs, any(jax.numpy.shape(lhs)), f"must be a scalar, not {type(lhs)}"
    )

    # Scale the components
    return replace(rhts, d_x=lhs * rhts.d_x, d_y=lhs * rhts.d_y)


#####################################################################


@final
class CartesianAcceleration2D(AvalMixin, AbstractAcceleration2D):
    """Cartesian acceleration representation."""

    d2_x: ct.BatchableSpeed = eqx.field(
        converter=partial(Quantity["acceleration"].constructor, dtype=float)
    )
    r"""X coordinate acceleration :math:`\frac{d^2 x}{dt^2} \in (-\infty,+\infty)`."""

    d2_y: ct.BatchableSpeed = eqx.field(
        converter=partial(Quantity["acceleration"].constructor, dtype=float)
    )
    r"""Y coordinate acceleration :math:`\frac{d^2 y}{dt^2} \in (-\infty,+\infty)`."""

    @classproperty
    @classmethod
    def integral_cls(cls) -> type[CartesianVelocity2D]:
        return CartesianVelocity2D

    # -----------------------------------------------------

    @override
    @partial(eqx.filter_jit, inline=True)
    def norm(self, _: AbstractVelocity2D | None = None, /) -> ct.BatchableAcc:
        """Return the norm of the vector.

        Examples
        --------
        >>> from unxt import Quantity
        >>> import coordinax as cx
        >>> v = cx.CartesianAcceleration2D.constructor([3, 4], "km/s2")
        >>> v.norm()
        Quantity['acceleration'](Array(5., dtype=float32), unit='km / s2')

        """
        return jnp.sqrt(self.d2_x**2 + self.d2_y**2)


# -----------------------------------------------------


@CartesianAcceleration2D.constructor._f.dispatch  # type: ignore[attr-defined, misc]  # noqa: SLF001
def constructor(
    cls: type[CartesianAcceleration2D],
    obj: AbstractQuantity,
    /,
) -> CartesianAcceleration2D:
    """Construct a 2D Cartesian velocity.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> vec = cx.CartesianAcceleration2D.constructor(Quantity([1, 2], "m/s2"))
    >>> vec
    CartesianAcceleration2D(
      d2_x=Quantity[...](value=f32[], unit=Unit("m / s2")),
      d2_y=Quantity[...](value=f32[], unit=Unit("m / s2"))
    )

    """
    comps = {f.name: obj[..., i] for i, f in enumerate(fields(cls))}
    return cls(**comps)


# -----------------------------------------------------


@register(jax.lax.add_p)  # type: ignore[misc]
def _add_aa(
    lhs: CartesianAcceleration2D, rhs: CartesianAcceleration2D, /
) -> CartesianAcceleration2D:
    """Add two Cartesian accelerations.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> v = cx.CartesianAcceleration2D.constructor(Quantity([3, 4], "km/s2"))
    >>> (v + v).d2_x
    Quantity['acceleration'](Array(6., dtype=float32), unit='km / s2')

    >>> jnp.add(v, v).d2_x
    Quantity['acceleration'](Array(6., dtype=float32), unit='km / s2')

    """
    return jax.tree.map(qlax.add, lhs, rhs)


@register(jax.lax.mul_p)  # type: ignore[misc]
def _mul_va(
    lhs: ArrayLike, rhts: CartesianAcceleration2D, /
) -> CartesianAcceleration2D:
    """Scale a cartesian 2D acceleration by a scalar.

    Examples
    --------
    >>> import quaxed.numpy as jnp
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> v = cx.CartesianAcceleration2D.constructor(Quantity([3, 4], "m/s2"))
    >>> jnp.multiply(5, v).d2_x
    Quantity['acceleration'](Array(15., dtype=float32), unit='m / s2')

    >>> (5 * v).d2_x
    Quantity['acceleration'](Array(15., dtype=float32), unit='m / s2')

    """
    # Validation
    lhs = eqx.error_if(
        lhs, any(jax.numpy.shape(lhs)), f"must be a scalar, not {type(lhs)}"
    )

    # Scale the components
    return replace(rhts, d2_x=lhs * rhts.d2_x, d2_y=lhs * rhts.d2_y)



"""Base classes for operators on coordinates and potentials."""

__all__ = ["AbstractCompositeOperator"]

from collections.abc import Iterator
from dataclasses import replace
from typing import TYPE_CHECKING, Protocol, overload, runtime_checkable

from dataclassish import DataclassInstance
from unxt import Quantity

from .base import AbstractOperator, op_call_dispatch
from coordinax._src.base import AbstractPosition

if TYPE_CHECKING:
    from typing_extensions import Self


@runtime_checkable
class HasOperatorsAttr(DataclassInstance, Protocol):  # type: ignore[misc]
    """Protocol for classes with an `operators` attribute."""

    operators: tuple[AbstractOperator, ...]


class AbstractCompositeOperator(AbstractOperator):
    """Abstract Composite Operator.

    This is the base class for all composite operations.

    See Also
    --------
    :class:`coordinax.operators.OperatorSequence`
    :class:`coordinax.operators.GalileanOperator`

    """

    # ===========================================
    # Operator

    # TODO: how to have the `operators` attribute in a way that allows for both
    # writeable (in the constructor) and read-only (as a property) subclasses.

    @op_call_dispatch(precedence=1)
    def __call__(
        self: "AbstractCompositeOperator", x: AbstractPosition, /
    ) -> AbstractPosition:
        """Apply the operator to the coordinates.

        This is the default implementation, which applies the operators in
        sequence.
        """
        # TODO: with lax.scan?
        for op in self.operators:
            x = op(x)
        return x

    @op_call_dispatch(precedence=1)
    def __call__(
        self: "AbstractCompositeOperator", x: AbstractPosition, t: Quantity["time"], /
    ) -> tuple[AbstractPosition, Quantity["time"]]:
        """Apply the operator to the coordinates."""
        # TODO: with lax.scan?
        for op in self.operators:
            x, t = op(x, t)
        return x, t

    @property
    def is_inertial(self: HasOperatorsAttr) -> bool:
        """Whether the operations maintain an inertial reference frame."""
        return all(op.is_inertial for op in self.operators)

    @property
    def inverse(self: HasOperatorsAttr) -> "AbstractCompositeOperator":
        """The inverse of the operator."""
        from .sequential import OperatorSequence

        return OperatorSequence(tuple(op.inverse for op in reversed(self.operators)))

    # ===========================================
    # Sequence

    @overload
    def __getitem__(self, key: int) -> AbstractOperator: ...

    @overload
    def __getitem__(self, key: slice) -> "Self": ...

    def __getitem__(self, key: int | slice) -> "AbstractOperator | Self":
        ops = self.operators[key]
        if isinstance(ops, AbstractOperator):
            return ops
        return replace(self, operators=ops)

    def __iter__(self: HasOperatorsAttr) -> Iterator[AbstractOperator]:
        return iter(self.operators)



"""Representation of coordinates in different systems."""

__all__ = ["AbstractAcceleration"]

from abc import abstractmethod
from functools import partial
from typing import TYPE_CHECKING, Any, TypeVar
from typing_extensions import override

import equinox as eqx
import jax
from quax import register

import quaxed.numpy as jnp
from dataclassish import field_items
from quaxed import lax as qlax
from unxt import Quantity

from .base import AbstractVector
from .base_pos import AbstractPosition
from .base_vel import AbstractVelocity
from coordinax._src.funcs import represent_as
from coordinax._src.utils import classproperty

if TYPE_CHECKING:
    from typing_extensions import Self

AccT = TypeVar("AccT", bound="AbstractAcceleration")

ACCELERATION_CLASSES: set[type["AbstractAcceleration"]] = set()


class AbstractAcceleration(AbstractVector):  # pylint: disable=abstract-method
    """Abstract representation of vector differentials in different systems."""

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Initialize the subclass.

        The subclass is registered.
        """
        ACCELERATION_CLASSES.add(cls)

    @classproperty
    @classmethod
    @abstractmethod
    def _cartesian_cls(cls) -> type["AbstractVector"]:
        """Return the corresponding Cartesian vector class.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.CartesianAcceleration3D._cartesian_cls
        <class 'coordinax...CartesianAcceleration3D'>

        >>> cx.SphericalAcceleration._cartesian_cls
        <class 'coordinax...CartesianAcceleration3D'>

        """
        # TODO: something nicer than this for getting the corresponding class
        raise NotImplementedError

    @classproperty
    @classmethod
    @abstractmethod
    def integral_cls(cls) -> type[AbstractVelocity]:
        """Return the corresponding vector class.

        Examples
        --------
        >>> import coordinax as cx

        >>> cx.RadialAcceleration.integral_cls.__name__
        'RadialVelocity'

        >>> cx.SphericalAcceleration.integral_cls.__name__
        'SphericalVelocity'

        """
        raise NotImplementedError

    # ===============================================================
    # Quax

    def aval(self) -> jax.core.ShapedArray:
        """Return the vector as a JAX array."""
        raise NotImplementedError

    # ===============================================================
    # Unary operations

    def __neg__(self) -> "Self":
        """Negate the vector.

        Examples
        --------
        >>> from unxt import Quantity
        >>> import coordinax as cx

        >>> d2r = cx.RadialAcceleration.constructor([1], "m/s2")
        >>> -d2r
        RadialAcceleration( d2_r=Quantity[...](value=i32[], unit=Unit("m / s2")) )

        >>> d2p = cx.PolarAcceleration(Quantity(1, "m/s2"), Quantity(1, "mas/yr2"))
        >>> negd2p = -d2p
        >>> negd2p.d2_r
        Quantity['acceleration'](Array(-1., dtype=float32), unit='m / s2')
        >>> negd2p.d2_phi
        Quantity['angular acceleration'](Array(-1., dtype=float32), unit='mas / yr2')

        """
        return jax.tree.map(jnp.negative, self)

    # ===============================================================
    # Convenience methods

    @override
    def represent_as(self, target: type[AccT], /, *args: Any, **kwargs: Any) -> AccT:
        """Represent the vector as another type.

        This just forwards to `coordinax.represent_as`.

        Parameters
        ----------
        target : type[`coordinax.AbstractVelocity`]
            The type to represent the vector as.
        *args, **kwargs : Any
            Extra arguments. These are passed to `coordinax.represent_as` and
            might be used, depending on the dispatched method. Generally the
            first argument is the velocity (`coordinax.AbstractVelocity`)
            followed by the position (`coordinax.AbstractPosition`) at which the
            acceleration is defined. In general this is a required argument,
            though it is not for Cartesian-to-Cartesian transforms -- see
            https://en.wikipedia.org/wiki/Tensors_in_curvilinear_coordinates for
            more information.

        Returns
        -------
        `coordinax.AbstractAcceleration`
            The vector represented as the target type.

        Examples
        --------
        >>> import coordinax as cx
        >>> q = cx.CartesianPosition3D.constructor([1, 2, 3], "m")
        >>> p = cx.CartesianVelocity3D.constructor([4, 5, 6], "m/s")
        >>> a = cx.CartesianAcceleration3D.constructor([7, 8, 9], "m/s2")
        >>> sph = a.represent_as(cx.SphericalAcceleration, p, q)
        >>> sph
        SphericalAcceleration(
            d2_r=Quantity[...](value=f32[], unit=Unit("m / s2")),
            d2_theta=Quantity[...]( value=f32[], unit=Unit("rad / s2") ),
            d2_phi=Quantity[...]( value=f32[], unit=Unit("rad / s2") )
        )
        >>> sph.d2_r
        Quantity['acceleration'](Array(13.363062, dtype=float32), unit='m / s2')

        """
        return represent_as(self, target, *args, **kwargs)

    @partial(eqx.filter_jit, inline=True)
    def norm(
        self, velocity: AbstractVelocity, position: AbstractPosition, /
    ) -> Quantity["speed"]:
        """Return the norm of the vector."""
        return self.represent_as(self._cartesian_cls, velocity, position).norm()


# ===============================================================


@register(jax.lax.mul_p)  # type: ignore[misc]
def _mul_acc_time(lhs: AbstractAcceleration, rhs: Quantity["time"]) -> AbstractVelocity:
    """Multiply the vector by a :class:`unxt.Quantity`.

    Examples
    --------
    >>> from quaxed import lax
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> d2r = cx.RadialAcceleration(Quantity(1, "m/s2"))
    >>> vec = lax.mul(d2r, Quantity(2, "s"))
    >>> vec
    RadialVelocity( d_r=Quantity[...]( value=...i32[], unit=Unit("m / s") ) )
    >>> vec.d_r
    Quantity['speed'](Array(2, dtype=int32, ...), unit='m / s')

    >>> (d2r * Quantity(2, "s")).d_r
    Quantity['speed'](Array(2, dtype=int32, ...), unit='m / s')

    """
    # TODO: better access to corresponding fields
    return lhs.integral_cls.constructor(
        {k.replace("2", ""): jnp.multiply(v, rhs) for k, v in field_items(lhs)}
    )


@register(jax.lax.mul_p)  # type: ignore[misc]
def _mul_time_acc(lhs: Quantity["time"], rhs: AbstractAcceleration) -> AbstractVelocity:
    """Multiply a scalar by an acceleration.

    Examples
    --------
    >>> from quaxed import lax
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> d2r = cx.RadialAcceleration(Quantity(1, "m/s2"))
    >>> vec = lax.mul(Quantity(2, "s"), d2r)
    >>> vec
    RadialVelocity( d_r=Quantity[...]( value=...i32[], unit=Unit("m / s") ) )
    >>> vec.d_r
    Quantity['speed'](Array(2, dtype=int32, ...), unit='m / s')

    """
    return qlax.mul(rhs, lhs)  # pylint: disable=arguments-out-of-order


@register(jax.lax.mul_p)  # type: ignore[misc]
def _mul_acc_time2(lhs: AbstractAcceleration, rhs: Quantity["s2"]) -> AbstractPosition:
    """Multiply an acceleration by a scalar.

    Examples
    --------
    >>> from quaxed import lax
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> d2r = cx.RadialAcceleration(Quantity(1, "m/s2"))
    >>> vec = lax.mul(d2r, Quantity(2, "s2"))
    >>> vec
    RadialPosition(r=Distance(value=f32[], unit=Unit("m")))
    >>> vec.r
    Distance(Array(2., dtype=float32), unit='m')

    >>> (d2r * Quantity(2, "s2")).r
    Distance(Array(2., dtype=float32), unit='m')

    """
    # TODO: better access to corresponding fields
    return lhs.integral_cls.integral_cls.constructor(
        {k.replace("d2_", ""): v * rhs for k, v in field_items(lhs)}
    )


@register(jax.lax.mul_p)  # type: ignore[misc]
def _mul_time2_acc(lhs: Quantity["s2"], rhs: AbstractAcceleration) -> AbstractPosition:
    """Multiply a scalar by an acceleration.

    Examples
    --------
    >>> from quaxed import lax
    >>> from unxt import Quantity
    >>> import coordinax as cx

    >>> d2r = cx.RadialAcceleration(Quantity(1, "m/s2"))
    >>> vec = lax.mul(Quantity(2, "s2"), d2r)
    >>> vec
    RadialPosition(r=Distance(value=f32[], unit=Unit("m")))
    >>> vec.r
    Distance(Array(2., dtype=float32), unit='m')

    """
    return qlax.mul(rhs, lhs)  # pylint: disable=arguments-out-of-order



"""Representation of coordinates in different systems."""

__all__ = ["Space"]

from collections.abc import Callable, ItemsView, Iterable, Mapping
from textwrap import indent
from types import MappingProxyType
from typing import Any, TypeAlias, final
from typing_extensions import override

import astropy.units as u
import equinox as eqx
import jax
from astropy.units import PhysicalType as Dimension
from jax import Device
from plum import dispatch

import quaxed.numpy as jnp
from unxt import Quantity, dimensions
from xmmutablemap import ImmutableMap

from .base import (
    AbstractAcceleration,
    AbstractPosition,
    AbstractVector,
    AbstractVelocity,
)
from .typing import Unit
from .utils import classproperty
from coordinax._src.funcs import represent_as

DimensionLike: TypeAlias = Dimension | str


def _get_dimension_name(dim: DimensionLike, /) -> str:
    return dimensions(dim)._physical_type_list[0]  # noqa: SLF001


def _can_broadcast_shapes(*shapes: tuple[int, ...]) -> bool:
    """Check if the shapes can be broadcasted together."""
    try:
        jnp.broadcast_shapes(*shapes)
    except ValueError:
        return False
    return True


# TODO: figure out how to make the keys into Dimension objects, not str. This is
#       running afoul of Jax's tree flattening, where ImmutableMap and
#       eqx.Module differ.
@final
class Space(ImmutableMap[Dimension, AbstractVector], AbstractVector):  # type: ignore[misc]
    """A collection of vectors that acts like the primary vector.

    Parameters
    ----------
    *args : Mapping[DimensionLike, AbstractVector] | tuple[DimensionLike, AbstractVector] | Iterable[tuple[DimensionLike, AbstractVector]]
        See input to `dict` for the input data.

    primary_key : DimensionLike, optional
        The key for the primary vector of the space.
        Default is "length" for position vectors.

    **kwargs : AbstractVector
        See input to `dict` for the input data.

    Examples
    --------
    >>> import coordinax as cx
    >>> from unxt import Quantity

    >>> x = cx.CartesianPosition3D.constructor([1, 2, 3], "km")
    >>> v = cx.CartesianVelocity3D.constructor([4, 5, 6], "km/s")
    >>> a = cx.CartesianAcceleration3D.constructor([7, 8, 9], "km/s^2")

    >>> space = cx.Space(length=x, speed=v, acceleration=a)
    >>> space
    Space({
        'length': CartesianPosition3D( ... ),
        'speed': CartesianVelocity3D( ... ),
        'acceleration': CartesianAcceleration3D( ... )}
    )

    >>> space["length"]
    CartesianPosition3D( ... )

    >>> space.represent_as(cx.SphericalPosition)
    Space({
        'length': SphericalPosition( ... ),
        'speed': SphericalVelocity( ... ),
        'acceleration': SphericalAcceleration( ... )}
    )

    >>> cx.represent_as(space, cx.SphericalPosition)
    Space({
        'length': SphericalPosition( ... ),
        'speed': SphericalVelocity( ... ),
        'acceleration': SphericalAcceleration( ... )}
    )

    """  # noqa: E501

    _data: dict[Dimension, AbstractVector]

    def __init__(  # pylint: disable=super-init-not-called  # TODO: resolve this
        self,
        /,
        *args: Mapping[DimensionLike, AbstractVector]
        | tuple[DimensionLike, AbstractVector]
        | Iterable[tuple[DimensionLike, AbstractVector]],
        **kwargs: AbstractVector,
    ) -> None:
        # Process the input data
        raw = dict(*args, **kwargs)  # process the input data
        keys = [_get_dimension_name(k) for k in raw]
        keys = eqx.error_if(
            keys,
            len(keys) < len(raw),
            f"Space(**input) contained duplicate keys {set(raw) - set(keys)}.",
        )
        # TODO: check the key dimension makes sense for the value

        # Check that the shapes are broadcastable
        keys = eqx.error_if(
            keys,
            not _can_broadcast_shapes(*(v.shape for v in raw.values())),
            "vector shapes are not broadcastable.",
        )

        super().__init__(dict(zip(keys, raw.values(), strict=True)))

    # ===============================================================
    # Mapping API

    @dispatch
    def __getitem__(self, key: Any) -> Any:
        """Get the vector by key.

        Examples
        --------
        >>> import coordinax as cx
        >>> w = cx.Space(length=cx.CartesianPosition3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...              speed=cx.CartesianVelocity3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m/s"))

        By number:

        >>> w[0]
        Space({
            'length': CartesianPosition3D(
                x=Quantity[...](value=f32[2], unit=Unit("m")),
                y=Quantity[...](value=f32[2], unit=Unit("m")),
                z=Quantity[...](value=f32[2], unit=Unit("m")) ),
            'speed': CartesianVelocity3D(
                d_x=Quantity[...]( value=f32[2], unit=Unit("m / s") ),
                d_y=Quantity[...]( value=f32[2], unit=Unit("m / s") ),
                d_z=Quantity[...]( value=f32[2], unit=Unit("m / s") ) )}
        )

        By slice:

        >>> w[1:]
        Space({
            'length': CartesianPosition3D(
                x=Quantity[...](value=f32[0,2], unit=Unit("m")),
                y=Quantity[...](value=f32[0,2], unit=Unit("m")),
                z=Quantity[...](value=f32[0,2], unit=Unit("m")) ),
            'speed': CartesianVelocity3D(
                d_x=Quantity[...]( value=f32[0,2], unit=Unit("m / s") ),
                d_y=Quantity[...]( value=f32[0,2], unit=Unit("m / s") ),
                d_z=Quantity[...]( value=f32[0,2], unit=Unit("m / s") ) )}
        )

        By Ellipsis:

        >>> w[...]
        Space({
            'length': CartesianPosition3D(
                x=Quantity[...](value=f32[1,2], unit=Unit("m")),
                y=Quantity[...](value=f32[1,2], unit=Unit("m")),
                z=Quantity[...](value=f32[1,2], unit=Unit("m")) ),
            'speed': CartesianVelocity3D(
                d_x=Quantity[...]( value=f32[1,2], unit=Unit("m / s") ),
                d_y=Quantity[...]( value=f32[1,2], unit=Unit("m / s") ),
                d_z=Quantity[...]( value=f32[1,2], unit=Unit("m / s") ) )}
        )

        By tuple[int, ...]:

        >>> w[(0, 1)]
        Space({
            'length': CartesianPosition3D(
                x=Quantity[...](value=f32[], unit=Unit("m")),
                y=Quantity[...](value=f32[], unit=Unit("m")),
                z=Quantity[...](value=f32[], unit=Unit("m")) ),
            'speed': CartesianVelocity3D(
                d_x=Quantity[...]( value=f32[], unit=Unit("m / s") ),
                d_y=Quantity[...]( value=f32[], unit=Unit("m / s") ),
                d_z=Quantity[...]( value=f32[], unit=Unit("m / s") ) )}
        )

        This also supports numpy index arrays. But this example section
        highlights core python indexing.

        """  # noqa: E501
        return Space(**{k: v[key] for k, v in self.items()})

    @dispatch
    def __getitem__(self, key: str | Dimension) -> Any:
        """Get the vector by key.

        Examples
        --------
        >>> import coordinax as cx
        >>> w = cx.Space(length=cx.CartesianPosition3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...              speed=cx.CartesianVelocity3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m/s"))

        By string key:

        >>> w["length"]
        CartesianPosition3D(
            x=Quantity[...](value=f32[1,2], unit=Unit("m")),
            y=Quantity[...](value=f32[1,2], unit=Unit("m")),
            z=Quantity[...](value=f32[1,2], unit=Unit("m"))
        )

        By dimension:

        >>> import astropy.units as u
        >>> w[u.get_physical_type("length")]
        CartesianPosition3D(
            x=Quantity[...](value=f32[1,2], unit=Unit("m")),
            y=Quantity[...](value=f32[1,2], unit=Unit("m")),
            z=Quantity[...](value=f32[1,2], unit=Unit("m"))
        )

        """  # noqa: E501
        if isinstance(key, Dimension):
            key = _get_dimension_name(key)

        return super().__getitem__(key)

    # ===============================================================
    # Quax

    def aval(self) -> jax.core.ShapedArray:
        """Return the vector as a JAX array."""
        raise NotImplementedError  # TODO: implement this

    # ===============================================================
    # Array API

    # ---------------------------------------------------------------
    # Attributes

    @property
    def mT(self) -> "Self":  # noqa: N802
        """Transpose each vector in the space.

        Examples
        --------
        >>> import coordinax as cx
        >>> from unxt import Quantity

        >>> w = cx.Space(
        ...     length=cx.CartesianPosition3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVelocity3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m/s")
        ... )
        >>> w.mT
        Space({
            'length': CartesianPosition3D(
                x=Quantity[...](value=f32[2,1], unit=Unit("m")),
                y=Quantity[...](value=f32[2,1], unit=Unit("m")),
                z=Quantity[...](value=f32[2,1], unit=Unit("m"))
            ), 'speed': CartesianVelocity3D(
                d_x=Quantity[...]( value=f32[2,1], unit=Unit("m / s") ),
                d_y=Quantity[...]( value=f32[2,1], unit=Unit("m / s") ),
                d_z=Quantity[...]( value=f32[2,1], unit=Unit("m / s") )
            )} )

        """  # noqa: E501
        return super().mT

    @property
    def ndim(self) -> int:
        """Number of array dimensions (axes).

        Examples
        --------
        >>> import coordinax as cx
        >>> from unxt import Quantity

        >>> w = cx.Space(
        ...     length=cx.CartesianPosition3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVelocity3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m/s") )

        >>> w.ndim
        2

        """  # noqa: E501
        return super().ndim

    @property
    def shape(self) -> tuple[int, ...]:
        """Get the shape of the vector's components.

        Examples
        --------
        >>> import coordinax as cx
        >>> from unxt import Quantity

        >>> w = cx.Space(
        ...     length=cx.CartesianPosition3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVelocity3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m/s")
        ... )

        >>> w.shape
        (1, 2)

        """  # noqa: E501
        return super().shape

    @property
    def size(self) -> int:
        """Total number of elements in the vector.

        Examples
        --------
        >>> import coordinax as cx
        >>> from unxt import Quantity

        >>> w = cx.Space(
        ...     length=cx.CartesianPosition3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVelocity3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m/s") )

        >>> w.size
        2

        """  # noqa: E501
        return super().size

    @property
    def T(self) -> "Self":  # noqa: N802
        """Transpose each vector in the space.

        Examples
        --------
        >>> import coordinax as cx
        >>> from unxt import Quantity

        >>> w = cx.Space(
        ...     length=cx.CartesianPosition3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVelocity3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m/s")
        ... )

        >>> w.T
        Space({
            'length': CartesianPosition3D(
                x=Quantity[...](value=f32[2,1], unit=Unit("m")),
                y=Quantity[...](value=f32[2,1], unit=Unit("m")),
                z=Quantity[...](value=f32[2,1], unit=Unit("m"))
            ),
            'speed': CartesianVelocity3D(
                d_x=Quantity[...]( value=f32[2,1], unit=Unit("m / s") ),
                d_y=Quantity[...]( value=f32[2,1], unit=Unit("m / s") ),
                d_z=Quantity[...]( value=f32[2,1], unit=Unit("m / s") )
            )} )

        """  # noqa: E501
        return super().T

    # ---------------------------------------------------------------
    # Methods

    def __neg__(self) -> "Self":
        """Negative of the vector.

        Examples
        --------
        >>> import coordinax as cx
        >>> from unxt import Quantity

        >>> w = cx.Space(
        ...     length=cx.CartesianPosition3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVelocity3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m/s")
        ... )

        >>> (-w)["length"].x
        Quantity['length'](Array([[-1., -4.]], dtype=float32), unit='m')

        """  # noqa: E501
        return type(self)(**{k: -v for k, v in self.items()})

    def __repr__(self) -> str:
        """Return the string representation.

        Examples
        --------
        >>> import coordinax as cx
        >>> from unxt import Quantity

        >>> q = cx.CartesianPosition3D.constructor([1, 2, 3], "m")
        >>> p = cx.CartesianVelocity3D.constructor([1, 2, 3], "m/s")
        >>> w = cx.Space(length=q, speed=p)
        >>> w
        Space({
            'length': CartesianPosition3D(
                x=Quantity[...](value=f32[], unit=Unit("m")),
                y=Quantity[...](value=f32[], unit=Unit("m")),
                z=Quantity[...](value=f32[], unit=Unit("m"))
            ),
            'speed': CartesianVelocity3D(
                d_x=Quantity[...]( value=f32[], unit=Unit("m / s") ),
                d_y=Quantity[...]( value=f32[], unit=Unit("m / s") ),
                d_z=Quantity[...]( value=f32[], unit=Unit("m / s") )
            )} )

        """
        cls_name = self.__class__.__name__
        data = "{\n" + indent(repr(self._data)[1:], "    ")
        return f"{cls_name}({data}\n)"

    def __str__(self) -> str:
        """Return the string representation."""
        return repr(self)

    # ===============================================================
    # Collection

    def asdict(
        self,
        *,
        dict_factory: Callable[[Any], Mapping[str, Quantity]] = dict,
    ) -> Mapping[str, Quantity]:
        """Return the vector as a Mapping.

        Parameters
        ----------
        dict_factory : type[Mapping]
            The type of the mapping to return.

        Returns
        -------
        Mapping[str, Quantity]
            The vector as a mapping.

        See Also
        --------
        `dataclasses.asdict`
            This applies recursively to the components of the vector.

        """
        return dict_factory(self._data)

    @override
    @classproperty
    @classmethod
    def components(cls) -> tuple[str, ...]:
        """Vector component names."""
        raise NotImplementedError  # TODO: implement this

    @property
    def units(self) -> MappingProxyType[str, Unit]:
        """Get the units of the vector's components."""
        raise NotImplementedError  # TODO: implement this

    @override
    @property
    def dtypes(self) -> MappingProxyType[str, MappingProxyType[str, jnp.dtype]]:
        """Get the dtypes of the vector's components.

        Examples
        --------
        >>> import coordinax as cx
        >>> from unxt import Quantity

        >>> w = cx.Space(
        ...     length=cx.CartesianPosition3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVelocity3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m/s")
        ... )

        >>> w.dtypes
        mappingproxy({'length': mappingproxy({'x': dtype('float32'), 'y': dtype('float32'), 'z': dtype('float32')}),
                      'speed': mappingproxy({'d_x': dtype('float32'), 'd_y': dtype('float32'), 'd_z': dtype('float32')})})

        """  # noqa: E501
        return MappingProxyType({k: v.dtypes for k, v in self.items()})

    @override
    @property
    def devices(self) -> MappingProxyType[str, MappingProxyType[str, Device]]:
        """Get the devices of the vector's components.

        Examples
        --------
        >>> import coordinax as cx
        >>> from unxt import Quantity

        >>> w = cx.Space(
        ...     length=cx.CartesianPosition3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVelocity3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m/s")
        ... )

        >>> w.devices
        mappingproxy({'length': mappingproxy({'x': CpuDevice(id=0), 'y': CpuDevice(id=0), 'z': CpuDevice(id=0)}),
                      'speed': mappingproxy({'d_x': CpuDevice(id=0), 'd_y': CpuDevice(id=0), 'd_z': CpuDevice(id=0)})})

        """  # noqa: E501
        return MappingProxyType({k: v.devices for k, v in self.items()})

    @override
    @property
    def shapes(self) -> MappingProxyType[str, MappingProxyType[str, tuple[int, ...]]]:  # type: ignore[override]
        """Get the shapes of the spaces's fields.

        Examples
        --------
        >>> import coordinax as cx
        >>> from unxt import Quantity

        >>> w = cx.Space(
        ...     length=cx.CartesianPosition3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVelocity3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m/s")
        ... )

        >>> w.shapes
        mappingproxy({'length': (1, 2), 'speed': (1, 2)})

        """  # noqa: E501
        return MappingProxyType({k: v.shape for k, v in self.items()})

    @property
    def sizes(self) -> MappingProxyType[str, int]:
        """Get the sizes of the vector's components.

        Examples
        --------
        >>> import coordinax as cx
        >>> from unxt import Quantity

        >>> w = cx.Space(
        ...     length=cx.CartesianPosition3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m"),
        ...     speed=cx.CartesianVelocity3D.constructor([[[1, 2, 3], [4, 5, 6]]], "m/s")
        ... )

        >>> w.sizes
        mappingproxy({'length': 2, 'speed': 2})

        """  # noqa: E501
        return MappingProxyType({k: v.size for k, v in self.items()})

    # ===============================================================
    # Convenience methods

    # TODO: correctly type annotate this method.
    def represent_as(self, target: type[AbstractVector], /) -> "Space":  # pylint: disable=W0221
        """Represent the vector as another type."""
        return represent_as(self, target)

    @dispatch  # type: ignore[misc]
    @override
    def to_units(
        self: "Space", units: Mapping[u.PhysicalType | str, Unit | str], /
    ) -> "Space":
        """Convert the vector to the given units."""
        raise NotImplementedError


# ===============================================================
# Related dispatches


@dispatch  # type: ignore[misc]
def represent_as(space: Space, target: type[AbstractVector], /) -> Space:
    """Represent the current vector to the target vector."""
    return type(space)(
        {k: temp_represent_as(v, target, space) for k, v in space.items()}
    )


# NOTE: need to set the precedence because `Space` is both a `Mapping` and a
#       `dataclass`, which are both in the `replace` dispatch table.
@dispatch(precedence=1)  # type: ignore[misc]
def replace(obj: Space, /, **kwargs: AbstractVector) -> Space:
    """Replace the components of the vector."""
    return type(obj)(**{**obj, **kwargs})


# NOTE: need to set the precedence because `Space` is both a `Mapping` and a
#       `dataclass`, which are both in the `field_items` dispatch table.
@dispatch(precedence=1)  # type: ignore[misc]
def field_items(obj: Space, /) -> ItemsView[str, AbstractVector]:
    """Return the items from a Space."""
    return obj.items()


# =============================================================== Temporary
# These functions are very similar to `represent_as`, but I don't think this is
# the best API. Until we figure out a better way to do this, we'll keep these
# functions here.


# TODO: should this be moved to a different file?
@dispatch
def temp_represent_as(
    current: AbstractPosition, target: type[AbstractPosition], space: Space, /
) -> AbstractPosition:
    """Transform of Positions."""
    return represent_as(current, target)  # space is unnecessary


# TODO: should this be moved to a different file?
@dispatch
def temp_represent_as(
    current: AbstractVelocity, target: type[AbstractPosition], space: Space, /
) -> AbstractVelocity:
    """Transform of Velocities."""
    return represent_as(current, target.differential_cls, space["length"])


# TODO: should this be moved to a different file?
@dispatch
def temp_represent_as(
    current: AbstractAcceleration, target: type[AbstractPosition], space: Space, /
) -> AbstractAcceleration:
    """Transform of Accelerations."""
    return represent_as(
        current,
        target.differential_cls.differential_cls,
        space["speed"],
        space["length"],
    )



"""Copyright (c) 2023 coordinax maintainers. All rights reserved."""

import os

RUNTIME_TYPECHECKER = (
    "beartype.beartype"
    if (os.environ.get("COORDINAX_ENABLE_RUNTIME_TYPECHECKS", "1") == "0")
    else None
)



"""Representation of coordinates in different systems."""

__all__: list[str] = []

from typing import Any

from plum import dispatch

from .base import AbstractAcceleration1D, AbstractPosition1D, AbstractVelocity1D
from .cartesian import CartesianAcceleration1D, CartesianPosition1D, CartesianVelocity1D
from .radial import RadialAcceleration, RadialPosition, RadialVelocity
from coordinax._src.base import AbstractPosition, AbstractVelocity

###############################################################################
# 1D


@dispatch
def represent_as(
    current: AbstractPosition1D, target: type[AbstractPosition1D], /, **kwargs: Any
) -> AbstractPosition1D:
    """AbstractPosition1D -> Cartesian1D -> AbstractPosition1D.

    This is the base case for the transformation of 1D vectors.
    """
    cart1d = represent_as(current, CartesianPosition1D)
    return represent_as(cart1d, target)


# TODO: use multi, with precedence
@dispatch(precedence=1)
def represent_as(
    current: CartesianPosition1D, target: type[CartesianPosition1D], /, **kwargs: Any
) -> CartesianPosition1D:
    """Self transform of 1D vectors."""
    return current


# TODO: use multi, with precedence
@dispatch(precedence=1)
def represent_as(
    current: RadialPosition, target: type[RadialPosition], /, **kwargs: Any
) -> RadialPosition:
    """Self transform of 1D vectors."""
    return current


@dispatch.multi(
    (CartesianVelocity1D, type[CartesianVelocity1D], AbstractPosition),
    (RadialVelocity, type[RadialVelocity], AbstractPosition),
)
def represent_as(
    current: AbstractVelocity1D,
    target: type[AbstractVelocity1D],
    position: AbstractPosition,
    /,
    **kwargs: Any,
) -> AbstractVelocity1D:
    """Self transform of 1D Velocities."""
    return current


# Special-case where self-transform doesn't need position
@dispatch.multi(
    (CartesianVelocity1D, type[CartesianVelocity1D]),
    (RadialVelocity, type[RadialVelocity]),
)
def represent_as(
    current: AbstractVelocity1D, target: type[AbstractVelocity1D], /, **kwargs: Any
) -> AbstractVelocity1D:
    """Self transform of 1D Velocities."""
    return current


@dispatch.multi(
    (
        CartesianAcceleration1D,
        type[CartesianAcceleration1D],
        AbstractVelocity,
        AbstractPosition,
    ),
    (RadialAcceleration, type[RadialAcceleration], AbstractVelocity, AbstractPosition),
)
def represent_as(
    current: AbstractAcceleration1D,
    target: type[AbstractAcceleration1D],
    velocity: AbstractVelocity,
    position: AbstractPosition,
    /,
    **kwargs: Any,
) -> AbstractAcceleration1D:
    """Self transform of 1D Accelerations."""
    return current


# Special-case where self-transform doesn't need velocity or position
@dispatch.multi(
    (CartesianAcceleration1D, type[CartesianAcceleration1D]),
    (RadialAcceleration, type[RadialAcceleration]),
)
def represent_as(
    current: AbstractAcceleration1D,
    target: type[AbstractAcceleration1D],
    /,
    **kwargs: Any,
) -> AbstractAcceleration1D:
    """Self transform of 1D Accelerations."""
    return current


# =============================================================================
# CartesianPosition1D


@dispatch
def represent_as(
    current: CartesianPosition1D, target: type[RadialPosition], /, **kwargs: Any
) -> RadialPosition:
    """CartesianPosition1D -> RadialPosition.

    The `x` coordinate is converted to the radial coordinate `r`.
    """
    return target(r=current.x)


# =============================================================================
# RadialPosition


@dispatch
def represent_as(
    current: RadialPosition, target: type[CartesianPosition1D], /, **kwargs: Any
) -> CartesianPosition1D:
    """RadialPosition -> CartesianPosition1D.

    The `r` coordinate is converted to the `x` coordinate of the 1D system.
    """
    return target(x=current.r.distance)


# =============================================================================
# CartesianVelocity1D


@dispatch
def represent_as(
    current: CartesianVelocity1D, target: type[CartesianVelocity1D], /
) -> CartesianVelocity1D:
    """CartesianVelocity1D -> CartesianVelocity1D with no position.

    Cartesian coordinates are an affine coordinate system and so the
    transformation of an n-th order derivative vector in this system do not
    require lower-order derivatives to be specified. See
    https://en.wikipedia.org/wiki/Tensors_in_curvilinear_coordinates for more
    information. This mixin provides a corresponding implementation of the
    `coordinax.represent_as` method for Cartesian velocities.

    Examples
    --------
    >>> import coordinax as cx
    >>> v = cx.CartesianVelocity1D.constructor([1], "m/s")
    >>> cx.represent_as(v, cx.CartesianVelocity1D) is v
    True

    """
    return current


# =============================================================================
# CartesianAcceleration1D


@dispatch
def represent_as(
    current: CartesianAcceleration1D, target: type[CartesianAcceleration1D], /
) -> CartesianAcceleration1D:
    """CartesianAcceleration1D -> CartesianAcceleration1D with no position.

    Cartesian coordinates are an affine coordinate system and so the
    transformation of an n-th order derivative vector in this system do not
    require lower-order derivatives to be specified. See
    https://en.wikipedia.org/wiki/Tensors_in_curvilinear_coordinates for more
    information. This mixin provides a corresponding implementation of the
    `coordinax.represent_as` method for Cartesian vectors.

    Examples
    --------
    >>> import coordinax as cx
    >>> a = cx.CartesianAcceleration1D.constructor([1], "m/s2")
    >>> cx.represent_as(a, cx.CartesianAcceleration1D) is a
    True

    """
    return current



"""Built-in vector classes."""

__all__ = [
    "TwoSpherePosition",
    "TwoSphereVelocity",
    "TwoSphereAcceleration",
]

from functools import partial
from typing import final
from typing_extensions import override

import equinox as eqx

from unxt import Quantity

import coordinax._src.typing as ct
from .base import AbstractAcceleration2D, AbstractPosition2D, AbstractVelocity2D
from coordinax._src.checks import check_azimuth_range, check_polar_range
from coordinax._src.converters import converter_azimuth_to_range
from coordinax._src.utils import classproperty


# TODO: should this be named TwoSphericalPosition or S(phere)(ical)2Position
@final
class TwoSpherePosition(AbstractPosition2D):
    r"""Position on the 2-Sphere.

    The space of coordinates on the unit sphere is called the 2-sphere or $S^2$.
    It is a two-dimensional surface embedded in three-dimensional space, defined
    by the set of all points at a unit distance from a central point.
    Mathematically, this is:

    $$ S^2 = \{ \mathbf{x} \in \mathbb{R}^3 |  \|\mathbf{x}\| = 1 \}. $$

    .. note::

        This class follows the Physics conventions (ISO 80000-2:2019).

    Parameters
    ----------
    theta : Quantity['angle']
        Polar angle [0, 180] [deg] where 0 is the z-axis.
    phi : Quantity['angle']
        Azimuthal angle [0, 360) [deg] where 0 is the x-axis.

    See Also
    --------
    `coordinax.SphericalPosition`
        The counterpart in $R^3$, adding the polar distance coordinate $r$.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    We can construct a 2-spherical coordinate:

    >>> s2 = cx.TwoSpherePosition(theta=Quantity(0, "deg"), phi=Quantity(180, "deg"))

    This coordinate has corresponding velocity class:

    >>> s2.differential_cls
    <class 'coordinax._src.d2.spherical.TwoSphereVelocity'>

    """

    theta: ct.BatchableAngle = eqx.field(
        converter=partial(Quantity["angle"].constructor, dtype=float)
    )
    r"""Inclination angle :math:`\theta \in [0,180]`."""

    phi: ct.BatchableAngle = eqx.field(
        converter=lambda x: converter_azimuth_to_range(
            Quantity["angle"].constructor(x, dtype=float)  # pylint: disable=E1120
        )
    )
    r"""Azimuthal angle :math:`\phi \in [0,360)`."""

    def __check_init__(self) -> None:
        """Check the validity of the initialization."""
        check_polar_range(self.theta)
        check_azimuth_range(self.phi)

    @override
    @classproperty
    @classmethod
    def differential_cls(cls) -> type["TwoSphereVelocity"]:
        return TwoSphereVelocity


#####################################################################


@final
class TwoSphereVelocity(AbstractVelocity2D):
    r"""Velocity on the 2-Sphere.

    The space of coordinates on the unit sphere is called the 2-sphere or $S^2$.
    It is a two-dimensional surface embedded in three-dimensional space, defined
    by the set of all points at a unit distance from a central point.
    Mathematically, this is:

    $$ S^2 = \{ \mathbf{x} \in \mathbb{R}^3 |  \|\mathbf{x}\| = 1 \}. $$

    .. note::

        This class follows the Physics conventions (ISO 80000-2:2019).

    Parameters
    ----------
    d_theta : Quantity['angular speed']
        Inclination speed $`d\theta/dt \in [-\infty, \infty]$.
    d_phi : Quantity['angular speed']
        Azimuthal speed $d\phi/dt \in [-\infty, \infty]$.

    See Also
    --------
    `coordinax.SphericalVelocity`
        The counterpart in $R^3$, adding the polar distance coordinate $d_r$.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    We can construct a 2-spherical velocity:

    >>> s2 = cx.TwoSphereVelocity(d_theta=Quantity(0, "deg/s"),
    ...                           d_phi=Quantity(2, "deg/s"))

    This coordinate has corresponding position and acceleration class:

    >>> s2.integral_cls
    <class 'coordinax._src.d2.spherical.TwoSpherePosition'>

    >>> s2.differential_cls
    <class 'coordinax._src.d2.spherical.TwoSphereAcceleration'>

    """

    d_theta: ct.BatchableAngularSpeed = eqx.field(
        converter=partial(Quantity["angular speed"].constructor, dtype=float)
    )
    r"""Inclination speed :math:`d\theta/dt \in [-\infty, \infty]."""

    d_phi: ct.BatchableAngularSpeed = eqx.field(
        converter=partial(Quantity["angular speed"].constructor, dtype=float)
    )
    r"""Azimuthal speed :math:`d\phi/dt \in [-\infty, \infty]."""

    @override
    @classproperty
    @classmethod
    def integral_cls(cls) -> type[TwoSpherePosition]:
        return TwoSpherePosition

    @override
    @classproperty
    @classmethod
    def differential_cls(cls) -> type["TwoSphereAcceleration"]:
        return TwoSphereAcceleration


@final
class TwoSphereAcceleration(AbstractAcceleration2D):
    r"""Velocity on the 2-Sphere.

    The space of coordinates on the unit sphere is called the 2-sphere or $S^2$.
    It is a two-dimensional surface embedded in three-dimensional space, defined
    by the set of all points at a unit distance from a central point.
    Mathematically, this is:

    $$ S^2 = \{ \mathbf{x} \in \mathbb{R}^3 |  \|\mathbf{x}\| = 1 \}. $$

    .. note::

        This class follows the Physics conventions (ISO 80000-2:2019).

    Parameters
    ----------
    d2_theta : Quantity['angular acceleration']
        Inclination acceleration $`d^2\theta/dt^2 \in [-\infty, \infty]$.
    d2_phi : Quantity['angular acceleration']
        Azimuthal acceleration $d^2\phi/dt^2 \in [-\infty, \infty]$.

    See Also
    --------
    `coordinax.SphericalAcceleration`
        The counterpart in $R^3$, adding the polar distance coordinate $d_r$.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax as cx

    We can construct a 2-spherical acceleration:

    >>> s2 = cx.TwoSphereAcceleration(d2_theta=Quantity(0, "deg/s2"),
    ...                               d2_phi=Quantity(2, "deg/s2"))

    This coordinate has corresponding velocity class:

    >>> s2.integral_cls
    <class 'coordinax._src.d2.spherical.TwoSphereVelocity'>

    """

    d2_theta: ct.BatchableAngularAcc = eqx.field(
        converter=partial(Quantity["angular acceleration"].constructor, dtype=float)
    )
    r"""Inclination acceleration :math:`d^2\theta/dt^2 \in [-\infty, \infty]."""

    d2_phi: ct.BatchableAngularAcc = eqx.field(
        converter=partial(Quantity["angular acceleration"].constructor, dtype=float)
    )
    r"""Azimuthal acceleration :math:`d^2\phi/dt^2 \in [-\infty, \infty]."""

    @override
    @classproperty
    @classmethod
    def integral_cls(cls) -> type[TwoSphereVelocity]:
        return TwoSphereVelocity



"""Sequence of Operators."""

__all__ = ["OperatorSequence"]

from dataclasses import replace
from typing import Any, final

import equinox as eqx

from .base import AbstractOperator
from .composite import AbstractCompositeOperator
from .funcs import simplify_op
from .identity import IdentityOperator


def _converter_seq(inp: Any) -> tuple[AbstractOperator, ...]:
    if isinstance(inp, tuple):
        return inp
    if isinstance(inp, OperatorSequence):
        return inp.operators
    if isinstance(inp, AbstractOperator):
        return (inp,)

    raise TypeError


@final
class OperatorSequence(AbstractCompositeOperator):
    """Sequence of operations.

    This is the composite operator that represents a sequence of operations to
    be applied in order.

    Parameters
    ----------
    operators : tuple[AbstractOperator, ...]
        The sequence of operators to apply.

    Examples
    --------
    >>> from unxt import Quantity
    >>> import coordinax.operators as co

    >>> shift = co.GalileanSpatialTranslationOperator(Quantity([1, 2, 3], "kpc"))
    >>> boost = co.GalileanBoostOperator(Quantity([10, 20, 30], "km/s"))
    >>> seq = co.OperatorSequence((shift, boost))
    >>> seq
    OperatorSequence(
      operators=( GalileanSpatialTranslationOperator( ... ),
                  GalileanBoostOperator( ... ) )
    )

    A sequence of operators can also be constructed by ``|``:

    >>> seq2 = shift | boost
    >>> seq2
    OperatorSequence(
      operators=( GalileanSpatialTranslationOperator( ... ),
                  GalileanBoostOperator( ... ) )
    )

    The sequence of operators can be simplified. For this example, we
    add an identity operator to the sequence:

    >>> seq3 = seq2 | co.IdentityOperator()
    >>> seq3
    OperatorSequence(
      operators=( GalileanSpatialTranslationOperator( ... ),
                  GalileanBoostOperator( ... ),
                  IdentityOperator() )
    )

    >>> co.simplify_op(seq3)
    OperatorSequence(
      operators=( GalileanSpatialTranslationOperator( ... ),
                  GalileanBoostOperator( ... ) )
    )

    """

    operators: tuple[AbstractOperator, ...] = eqx.field(converter=_converter_seq)

    def __or__(self, other: AbstractOperator) -> "OperatorSequence":
        """Compose with another operator."""
        if isinstance(other, type(self)):
            return replace(self, operators=(*self, *other.operators))
        return replace(self, operators=(*self, other))

    def __ror__(self, other: AbstractOperator) -> "OperatorSequence":
        return replace(self, operators=(other, *self))


#####################################################################
# Functions


@simplify_op.register
def _simplify_op(seq: OperatorSequence, /) -> OperatorSequence:
    """Simplify a sequence of Operators.

    This simplifies the sequence of operators by removing any that reduce to
    the Identity operator.
    """
    # Iterate through the operators, simplifying that operator, then filtering
    # out any that reduce to the Identity.
    # TODO: this doesn't do any type of operator fusion, e.g. a
    # ``GalileanRotationOperator | GalileanTranslationOperator |
    # GalileanBoostOperator => GalileanOperator``
    return OperatorSequence(
        tuple(
            sop
            for op in seq.operators
            if not isinstance((sop := simplify_op(op)), IdentityOperator)
        )
    )



"""Representation of coordinates in different systems."""

__all__: list[str] = []

from typing import TypeAlias

import astropy.units as u
from jaxtyping import Float, Int, Shaped

from unxt import AbstractDistance, AbstractQuantity, Quantity

Unit: TypeAlias = u.Unit | u.UnitBase | u.CompositeUnit

FloatScalarQ = Float[AbstractQuantity, ""]
BatchFloatScalarQ = Shaped[FloatScalarQ, "*batch"]
BatchableFloatScalarQ = Shaped[FloatScalarQ, "*#batch"]

ScalarTime = Float[Quantity["time"], ""] | Int[Quantity["time"], ""]

BatchableAngle = Shaped[Quantity["angle"], "*#batch"]
BatchableLength = Shaped[Quantity["length"], "*#batch"]
BatchableDistance = Shaped[AbstractDistance, "*#batch"]
BatchableTime = Shaped[Quantity["time"], "*#batch"]

BatchableSpeed = Shaped[Quantity["speed"], "*#batch"]
BatchableAngularSpeed = Shaped[Quantity["angular speed"], "*#batch"]

BatchableAcc = Shaped[Quantity["acceleration"], "*#batch"]
BatchableAngularAcc = Shaped[Quantity["angular acceleration"], "*#batch"]


TimeBatchOrScalar = ScalarTime | BatchableTime
